import collections
import hashlib
from typing import Callable, Dict, List, Literal, Optional, Tuple, Union

import imutils
import numpy as np
from PIL import Image, ImageColor, ImageDraw, ImageFont

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.utils.images import rotate_point
from cv_pipeliner.utils.images_datas import (
    flatten_additional_bboxes_data_in_image_data,
    get_image_data_filtered_by_labels,
)

# Taken from object_detection.utils.visualization_utils
STANDARD_COLORS = [
    "AliceBlue",
    "Chartreuse",
    "Aqua",
    "Aquamarine",
    "Azure",
    "Beige",
    "Bisque",
    "BlanchedAlmond",
    "BlueViolet",
    "BurlyWood",
    "CadetBlue",
    "AntiqueWhite",
    "Chocolate",
    "Coral",
    "CornflowerBlue",
    "Cornsilk",
    "Crimson",
    "Cyan",
    "DarkCyan",
    "DarkGoldenRod",
    "DarkGrey",
    "DarkKhaki",
    "DarkOrange",
    "DarkOrchid",
    "DarkSalmon",
    "DarkSeaGreen",
    "DarkTurquoise",
    "DarkViolet",
    "DeepPink",
    "DeepSkyBlue",
    "DodgerBlue",
    "FireBrick",
    "FloralWhite",
    "ForestGreen",
    "Fuchsia",
    "Gainsboro",
    "GhostWhite",
    "Gold",
    "GoldenRod",
    "Salmon",
    "Tan",
    "HoneyDew",
    "HotPink",
    "IndianRed",
    "Ivory",
    "Khaki",
    "Lavender",
    "LavenderBlush",
    "LawnGreen",
    "LemonChiffon",
    "LightBlue",
    "LightCoral",
    "LightCyan",
    "LightGoldenRodYellow",
    "LightGray",
    "LightGrey",
    "LightGreen",
    "LightPink",
    "LightSalmon",
    "LightSeaGreen",
    "LightSkyBlue",
    "LightSlateGray",
    "LightSlateGrey",
    "LightSteelBlue",
    "LightYellow",
    "Lime",
    "LimeGreen",
    "Linen",
    "Magenta",
    "MediumAquaMarine",
    "MediumOrchid",
    "MediumPurple",
    "MediumSeaGreen",
    "MediumSlateBlue",
    "MediumSpringGreen",
    "MediumTurquoise",
    "MediumVioletRed",
    "MintCream",
    "MistyRose",
    "Moccasin",
    "NavajoWhite",
    "OldLace",
    "Olive",
    "OliveDrab",
    "Orange",
    "OrangeRed",
    "Orchid",
    "PaleGoldenRod",
    "PaleGreen",
    "PaleTurquoise",
    "PaleVioletRed",
    "PapayaWhip",
    "PeachPuff",
    "Peru",
    "Pink",
    "Plum",
    "PowderBlue",
    "Purple",
    "Red",
    "RosyBrown",
    "RoyalBlue",
    "SaddleBrown",
    "Green",
    "SandyBrown",
    "SeaGreen",
    "SeaShell",
    "Sienna",
    "Silver",
    "SkyBlue",
    "SlateBlue",
    "SlateGray",
    "SlateGrey",
    "Snow",
    "SpringGreen",
    "SteelBlue",
    "GreenYellow",
    "Teal",
    "Thistle",
    "Tomato",
    "Turquoise",
    "Violet",
    "Wheat",
    "White",
    "WhiteSmoke",
    "Yellow",
    "YellowGreen",
]
STANDARD_COLORS_RGB = ["Red", "Green", "Blue", "Yellow"]
STANDARD_COLORS_RGB = STANDARD_COLORS_RGB + [x for x in STANDARD_COLORS if x not in STANDARD_COLORS_RGB]


def draw_mask_on_image(image: np.ndarray, mask: np.ndarray, color: Tuple[int, int, int], alpha: float) -> np.ndarray:
    if len(mask.shape) == 2:
        mask = np.array(Image.fromarray(mask).convert("RGB"))
    norm_mask = mask.astype(np.float32) / 255.0
    np_where = np.where(norm_mask > 0)
    if len(np_where[0]) > 0:
        colored_mask_np = np.zeros_like(mask, dtype=np.float32)
        recolor_value = np.array(color)
        colored_mask_np[np_where] = norm_mask[np_where] * np.tile(recolor_value, len(np_where[0]) // 3)
        colored_mask = colored_mask_np.astype(np.uint8)
        colored_np_where = np.where(colored_mask > 0)
        image[colored_np_where] = (1 - alpha) * image[colored_np_where] + alpha * colored_mask[colored_np_where]
    return image


# Taken from object_detection.utils.visualization_utils
def draw_bounding_box_on_image(
    image: Image,
    xmin: int,
    ymin: int,
    xmax: int,
    ymax: int,
    keypoints: np.ndarray = [],
    angle: int = 0,
    color="red",
    thickness=4,
    display_str_list=(),
    use_normalized_coordinates=True,
    keypoints_radius: int = 5,
    fontsize: int = 24,
):
    """Adds a bounding box to an image.

    Bounding box coordinates can be specified in either absolute (pixel) or
    normalized coordinates by setting the use_normalized_coordinates argument.

    Each string in display_str_list is displayed on a separate line above the
    bounding box in black text on a rectangle filled with the input 'color'.
    If the top of the bounding box extends to the edge of the image, the strings
    are displayed below the bounding box.

    Args:
    image: a PIL.Image object.
    ymin: ymin of bounding box.
    xmin: xmin of bounding box.
    ymax: ymax of bounding box.
    xmax: xmax of bounding box.
    angle: angle of bounding box.
    color: color to draw bounding box. Default is red.
    thickness: line thickness. Default value is 4.
    display_str_list: list of strings to display in box
                        (each to be shown on its own line).
    use_normalized_coordinates: If True (default), treat coordinates
        ymin, xmin, ymax, xmax as relative to the image.  Otherwise treat
        coordinates as absolute.
    """
    draw = ImageDraw.Draw(image)
    im_width, im_height = image.size
    if use_normalized_coordinates:
        (left, right, top, bottom) = (xmin * im_width, xmax * im_width, ymin * im_height, ymax * im_height)
    else:
        (left, right, top, bottom) = (xmin, xmax, ymin, ymax)
    points = [(left, top), (left, bottom), (right, bottom), (right, top), (left, top)]
    rotated_points = [rotate_point(x=x, y=y, cx=left, cy=top, angle=angle) for (x, y) in points]
    if thickness > 0:
        draw.line(rotated_points, width=thickness, fill=color)
    try:
        font = ImageFont.truetype("arial.ttf", fontsize)
    except IOError:
        font = ImageFont.load_default()

    # If the total height of the display strings added to the top of the bounding
    # box exceeds the top of the image, stack the strings below the bounding box
    # instead of above.
    display_str_bboxes = [font.getbbox(ds) for ds in display_str_list]
    display_str_heights = [f_ymax - f_ymin for (_, f_ymin, _, f_ymax) in display_str_bboxes]
    # Each display_str has a top and bottom margin of 0.05x.
    total_display_str_height = (1 + 2 * 0.05) * sum(display_str_heights)

    if top > total_display_str_height:
        text_bottom = top
    else:
        text_bottom = bottom + total_display_str_height
    # Reverse list and print from bottom to top.
    for display_str in display_str_list[::-1]:
        f_xmin, f_ymin, f_xmax, f_ymax = font.getbbox(display_str)
        text_width, text_height = f_xmax - f_xmin, f_ymax - f_ymin
        margin = np.ceil(0.05 * text_height)
        draw.rectangle([(left, text_bottom - text_height - 2 * margin), (left + text_width, text_bottom)], fill=color)
        draw.text((left + margin, text_bottom - text_height - margin), display_str, fill="black", font=font)
        text_bottom -= text_height - 2 * margin

    keypoints = np.array(keypoints).reshape(len(keypoints), 2)
    for idx, (x, y) in enumerate(keypoints):
        draw.pieslice(
            [(x - keypoints_radius, y - keypoints_radius), (x + keypoints_radius, y + keypoints_radius)],
            start=0,
            end=360,
            fill=STANDARD_COLORS_RGB[idx % len(STANDARD_COLORS_RGB)],
        )


# Taken from object_detection.utils.visualization_utils
def visualize_boxes_and_labels_on_image_array(
    image: np.ndarray,
    bboxes: List[Tuple[int, int, int, int]],
    angles: List[int],
    labels: List[str],
    scores: List[float],
    k_keypoints: List[List[Tuple[int, int]]],
    use_normalized_coordinates=False,
    groundtruth_box_visualization_color="black",
    known_labels: List[str] = [],
    skip_scores=False,
    skip_labels=False,
    keypoints_radius: int = 5,
    fontsize: int = 24,
    thickness: int = 4,
    label_to_color: Optional[Dict[str, str]] = None,
):
    """Overlay labeled boxes on an image with formatted scores and label names.

    This function groups boxes that correspond to the same location
    and creates a display string for each detection and overlays these
    on the image. Note that this function modifies the image in place, and returns
    that same image.

    Args:
      image: uint8 numpy array with shape (img_height, img_width, 3)
      boxes: a numpy array of shape [N, 4]
      angles: a numpy array of shape [N].
      labels: a numpy array of shape [N]. Note that class indices are 1-based.
      scores: a numpy array of shape [N] or None.  If scores=None, then
        this function assumes that the boxes to be plotted are groundtruth
        boxes and plot all boxes as black with no classes or scores.
      use_normalized_coordinates: whether boxes is to be interpreted as
        normalized coordinates or not.
      line_thickness: integer (default: 4) controlling line width of the boxes.
      groundtruth_box_visualization_color: box color for visualizing groundtruth
        boxes
      known_labels: a list of known labels. If given, bboxes colors will be chosen by this list.
      skip_boxes: whether to skip the drawing of bounding boxes.
      skip_scores: whether to skip score when drawing a single detection
      skip_labels: whether to skip label when drawing a single detection

    Returns:
      uint8 numpy array with shape (img_height, img_width, 3) with overlaid boxes.
    """
    bboxes = np.array(bboxes)
    labels = np.array(labels)
    scores = np.array(scores)

    if len(known_labels) > 0:
        known_labels = set(known_labels)
    label_to_id = {label: int(hashlib.md5(str(label).encode()).hexdigest(), 16) for id_, label in enumerate(labels)}
    bbox_to_display_str = collections.defaultdict(list)
    bbox_to_color = collections.defaultdict(str)

    for i in range(len(bboxes)):
        bbox = tuple(bboxes[i].tolist())
        if skip_labels and len(known_labels) == 0:
            bbox_to_color[bbox] = groundtruth_box_visualization_color
        else:
            display_str = ""
            if not skip_labels:
                display_str = str(labels[i])
            if not skip_scores:
                if not display_str:
                    display_str = f"{round(100*scores[i])}%"
                else:
                    display_str = f"{display_str}: {round(100*scores[i])}%"
            bbox_to_display_str[bbox].append(display_str)
            if len(known_labels) > 0 and labels[i] in known_labels:
                if label_to_color is not None:
                    bbox_to_color[bbox] = label_to_color[labels[i]]
                else:
                    bbox_to_color[bbox] = STANDARD_COLORS[label_to_id[labels[i]] % len(STANDARD_COLORS)]
            else:
                bbox_to_color[bbox] = groundtruth_box_visualization_color

    # Draw all boxes onto image.
    image_pil = Image.fromarray(np.uint8(image)).convert("RGB")
    for bbox, angle, keypoints in zip(bboxes, angles, k_keypoints):
        bbox = tuple(bbox.tolist())
        xmin, ymin, xmax, ymax = bbox
        draw_bounding_box_on_image(
            image=image_pil,
            xmin=xmin,
            ymin=ymin,
            xmax=xmax,
            ymax=ymax,
            keypoints=keypoints,
            angle=angle,
            color=bbox_to_color[bbox],
            thickness=thickness,
            display_str_list=bbox_to_display_str[bbox],
            use_normalized_coordinates=use_normalized_coordinates,
            keypoints_radius=keypoints_radius,
            fontsize=fontsize,
        )
    image = np.array(image_pil)
    return image


def draw_label_image(
    image: np.ndarray, base_label_image: np.ndarray, bbox_data: BboxData, inplace: bool = False
) -> np.ndarray:
    if not inplace:
        image = image.copy()

    bbox_data_size = max(bbox_data.xmax - bbox_data.xmin, bbox_data.ymax - bbox_data.ymin)
    resize = min(int(bbox_data_size / 1.5), int(max(image.shape) / 20))

    height, width, _ = base_label_image.shape
    if height <= width:
        label_image = imutils.resize(base_label_image, width=resize)
    else:
        label_image = imutils.resize(base_label_image, height=resize)

    x_offset = bbox_data.xmin - 20
    y_offset = bbox_data.ymax - label_image.shape[0]

    y_min, y_max = y_offset, y_offset + label_image.shape[0]
    x_min, x_max = x_offset, x_offset + label_image.shape[1]

    # Ensure that label image is inside image boundaries
    if y_max > image.shape[0]:
        y_min -= y_max - image.shape[0]
        y_max = image.shape[0]

    if x_max > image.shape[1]:
        x_min -= x_max - image.shape[1]
        x_max = image.shape[1]

    if x_min < 0:
        x_max -= x_min
        x_min = 0

    if y_min < 0:
        y_max -= y_min
        y_min = 0

    alpha_label_image = label_image[:, :, 3] / 255.0
    alpha_image = 1.0 - alpha_label_image

    for channel in range(0, 3):
        image[y_min:y_max, x_min:x_max, channel] = (
            alpha_label_image * label_image[:, :, channel] + alpha_image * image[y_min:y_max, x_min:x_max, channel]
        )

    if not inplace:
        return image


def visualize_image_data(
    image_data: ImageData,
    include_labels: bool = False,
    score_type: Literal["detection", "classification"] = None,
    filter_by_labels: List[str] = None,
    known_labels: Optional[List[str]] = None,
    draw_base_labels_with_given_label_to_base_label_image: Callable[[str], np.ndarray] = None,
    keypoints_radius: int = 5,
    include_additional_bboxes_data: bool = False,
    additional_bboxes_data_depth: Optional[int] = None,
    include_keypoints: bool = False,
    include_mask: bool = False,
    mask_alpha: float = 0.5,
    label_to_color: Optional[Dict[str, str]] = None,
    fontsize: int = 24,
    thickness: int = 4,
    thumbnail_size: Optional[Union[int, Tuple[int, int]]] = None,
    return_as_pil_image: bool = False,
    default_color: str = "lime",
    exif_transpose: bool = False,
    xmin_offset: Union[int, float] = 0,
    ymin_offset: Union[int, float] = 0,
    xmax_offset: Union[int, float] = 0,
    ymax_offset: Union[int, float] = 0,
    use_labels: Optional[bool] = None,
) -> Union[np.ndarray, Image.Image]:
    if use_labels is not None:
        print("WARNING: argument use_labels= is deprecated and will be removed. Use include_labels= instead.")
    if thumbnail_size is not None:
        from cv_pipeliner.utils.images_datas import thumbnail_image_data

        image_data = thumbnail_image_data(image_data, thumbnail_size)

    image_data = get_image_data_filtered_by_labels(image_data=image_data, filter_by_labels=filter_by_labels)
    image = image_data.open_image(exif_transpose=exif_transpose)
    if include_additional_bboxes_data:
        bboxes_data = flatten_additional_bboxes_data_in_image_data(
            image_data, additional_bboxes_data_depth=additional_bboxes_data_depth
        ).bboxes_data
    else:
        bboxes_data = image_data.bboxes_data
    labels = [bbox_data.label for bbox_data in bboxes_data] + [image_data.label]
    if known_labels is None:
        known_labels = list(set(labels))
    k_keypoints = [bbox_data.keypoints for bbox_data in bboxes_data]
    bboxes = np.array(
        [
            bbox_data.coords_with_offset(xmin_offset, ymin_offset, xmax_offset, ymax_offset, source_image=image)
            for bbox_data in bboxes_data
        ]
    )
    angles = np.array([0.0 for _ in bboxes_data])
    if score_type == "detection":
        scores = np.array([bbox_data.detection_score for bbox_data in bboxes_data])
        skip_scores = False
    elif score_type == "classification":
        scores = np.array([bbox_data.classification_score for bbox_data in bboxes_data])
        skip_scores = False
    else:
        scores = None
        skip_scores = True

    image = visualize_boxes_and_labels_on_image_array(
        image=image,
        bboxes=bboxes,
        angles=angles,
        scores=scores,
        k_keypoints=k_keypoints,
        labels=labels,
        use_normalized_coordinates=False,
        skip_scores=skip_scores,
        skip_labels=not include_labels,
        groundtruth_box_visualization_color=default_color,
        known_labels=known_labels,
        keypoints_radius=keypoints_radius,
        fontsize=fontsize,
        thickness=thickness,
        label_to_color=label_to_color,
    )
    if include_keypoints and len(image_data.keypoints) > 0:
        image_pil = Image.fromarray(image)
        draw = ImageDraw.Draw(image_pil)
        for idx, (x, y) in enumerate(image_data.keypoints):
            draw.pieslice(
                [(x - keypoints_radius, y - keypoints_radius), (x + keypoints_radius, y + keypoints_radius)],
                start=0,
                end=360,
                fill=STANDARD_COLORS_RGB[idx % len(STANDARD_COLORS_RGB)],
            )
        image = np.array(image_pil)

    if include_mask:
        label_to_id = {label: int(hashlib.md5(str(label).encode()).hexdigest(), 16) for id_, label in enumerate(labels)}
        for bbox_data in bboxes_data:
            if label_to_color is None:
                bbox_color = ImageColor.getrgb(STANDARD_COLORS[label_to_id[bbox_data.label] % len(STANDARD_COLORS)])
            else:
                bbox_color = ImageColor.getrgb(label_to_color.get(bbox_data.label, default_color))
            bbox_mask_np = bbox_data.open_mask(exif_transpose=exif_transpose)
            image = draw_mask_on_image(
                image=image,
                mask=bbox_mask_np,
                color=bbox_color,
                alpha=mask_alpha,
            )
        mask_np = image_data.open_mask(exif_transpose=exif_transpose, include_bboxes_data=False)
        if label_to_color is None:
            color = ImageColor.getrgb(STANDARD_COLORS[label_to_id[image_data.label] % len(STANDARD_COLORS)])
        else:
            color = ImageColor.getrgb(label_to_color.get(image_data.label, default_color))
        image = draw_mask_on_image(
            image=image,
            mask=mask_np,
            color=color,
            alpha=mask_alpha,
        )

    if draw_base_labels_with_given_label_to_base_label_image is not None:
        for bbox_data in bboxes_data:
            base_label_image = draw_base_labels_with_given_label_to_base_label_image(bbox_data.label)
            draw_label_image(image=image, base_label_image=base_label_image, bbox_data=bbox_data, inplace=True)

    if return_as_pil_image:
        return Image.fromarray(image)

    return image



from dataclasses import dataclass
from typing import List, Literal

import numpy as np

from cv_pipeliner.core.data import BboxData, ImageData


def intersection_over_union(bbox_data1: BboxData, bbox_data2: BboxData) -> float:
    if bbox_data1 is None or bbox_data2 is None:
        return None
    xmins = np.max([bbox_data1.xmin, bbox_data2.xmin])
    ymins = np.max([bbox_data1.ymin, bbox_data2.ymin])
    xmaxs = np.min([bbox_data1.xmax, bbox_data2.xmax])
    ymaxs = np.min([bbox_data1.ymax, bbox_data2.ymax])
    inter_area = np.max([0, ymaxs - ymins + 1]) * np.max([0, xmaxs - xmins + 1])
    bbox1_area = (bbox_data1.xmax - bbox_data1.xmin + 1) * (bbox_data1.ymax - bbox_data1.ymin + 1)
    bbox2_area = (bbox_data2.xmax - bbox_data2.xmin + 1) * (bbox_data2.ymax - bbox_data2.ymin + 1)
    iou = inter_area / float(bbox1_area + bbox2_area - inter_area + 1e-9)
    return iou


def pairwise_intersection_over_union(
    bboxes_data1: List[BboxData], bboxes_data2: List[BboxData]  # N  # K
) -> List[List[float]]:  # Matrix NxK
    low = np.s_[..., :2]
    high = np.s_[..., 2:]
    A = np.array([[bbox_data.xmin, bbox_data.ymin, bbox_data.xmax, bbox_data.ymax] for bbox_data in bboxes_data1])
    B = np.array([[bbox_data.xmin, bbox_data.ymin, bbox_data.xmax, bbox_data.ymax] for bbox_data in bboxes_data2])
    A = A[:, None]
    B = B[None]
    A[high] += 1
    B[high] += 1
    interactions = np.prod(np.maximum(0, np.minimum(A[high], B[high]) - np.maximum(A[low], B[low])), axis=-1)
    bboxes_areas1 = np.prod(A[high] - A[low], axis=-1)
    bboxes_areas2 = np.prod(B[high] - B[low], axis=-1)
    return interactions / (bboxes_areas1 + bboxes_areas2 - interactions + 1e-9)


@dataclass
class BboxDataMatching:
    """
    A dataclass providing macthing between true_bbox_data and pred_bbox_data inside of given image_data.
    By using this matching, we can get error type (TP, TN, FP, FN) for Detection or Pipeline models.
    """

    true_bbox_data: BboxData = None
    pred_bbox_data: BboxData = None
    extra_bbox_label: str = None

    def __post_init__(self):
        if self.extra_bbox_label is None:
            self.extra_bbox_label = "trash"
        self._iou = intersection_over_union(self.true_bbox_data, self.pred_bbox_data)

    def get_detection_error_type(self, label: str = None) -> Literal["TP", "TN", "FP", "FN"]:
        if label is not None and self.true_bbox_data is not None and self.true_bbox_data.label != label:
            return "TN"
        # true_bbox is found:
        if self.true_bbox_data is not None and self.pred_bbox_data is not None:
            return "TP"
        # true_bbox is not found:
        elif self.true_bbox_data is not None and self.pred_bbox_data is None:
            return "FN"
        # pred_bbox is an extra:
        elif self.true_bbox_data is None and self.pred_bbox_data is not None:
            return "FP"

    def get_pipeline_error_type(
        self, label: str = None
    ) -> Literal["TP", "TN", "FP", "FN", "TP (extra bbox)", "TN (extra bbox)", "FP (extra bbox)", "FN (extra bbox)"]:
        for bbox_data in [self.true_bbox_data, self.pred_bbox_data]:
            if bbox_data is not None:
                assert bbox_data.label is not None

        true_label = self.true_bbox_data.label if self.true_bbox_data is not None else None
        pred_label = self.pred_bbox_data.label if self.pred_bbox_data is not None else None

        assert true_label is not None or pred_label is not None

        # true_bbox is found:
        if self.true_bbox_data is not None and self.pred_bbox_data is not None:
            if label is None:
                if true_label != pred_label:
                    return "FP"
                else:
                    return "TP"
            else:
                if true_label != label and pred_label != label:
                    return "TN"
                elif true_label == label and pred_label == label:
                    return "TP"
                elif true_label != label and pred_label == label:
                    return "FP"
                else:
                    return "FN"
        # true_bbox is not found:
        elif self.true_bbox_data is not None and self.pred_bbox_data is None:
            if label is None:
                return "FN"
            else:
                if true_label == label:
                    return "FN"
                else:
                    return "TN"
        # pred_bbox is an extra
        elif self.true_bbox_data is None and self.pred_bbox_data is not None:
            if label is None:
                if pred_label != self.extra_bbox_label:
                    return "FP (extra bbox)"
                else:
                    return "TP (extra bbox)"
            else:
                if label != self.extra_bbox_label and pred_label != label:
                    return "TN (extra bbox)"
                elif label == self.extra_bbox_label and pred_label == label:
                    return "TP (extra bbox)"
                elif label != self.extra_bbox_label and pred_label == label:
                    return "FP (extra bbox)"
                else:
                    return "FN (extra bbox)"

    @property
    def iou(self):
        return self._iou


@dataclass(init=False)
class ImageDataMatching:
    """
    A dataclass providing macthing between true_bboxes_data and pred_bboxes_data
    inside of given image_data.

    We say that pred_bbox_data is matched to true_bbox_data if they have iou >= minimum_iou.
    One true_bbox_data may have only one matching to the pred_bbox_data.

    For illustrations, look tests/test_image_data_matching.py
    """

    true_image_data: ImageData
    pred_image_data: ImageData
    minimum_iou: float
    extra_bbox_label: str
    bboxes_data_matchings: List[BboxDataMatching]

    def __init__(
        self,
        true_image_data: ImageData,
        pred_image_data: ImageData,
        minimum_iou: float,
        extra_bbox_label: str = None,
        bboxes_data_matchings: List[BboxDataMatching] = None,
    ):
        self.true_image_data = true_image_data
        self.pred_image_data = pred_image_data
        self.minimum_iou = minimum_iou
        self.extra_bbox_label = extra_bbox_label
        if bboxes_data_matchings is None:
            self.bboxes_data_matchings = self._get_bboxes_data_matchings(
                true_image_data=true_image_data,
                pred_image_data=pred_image_data,
                minimum_iou=minimum_iou,
                extra_bbox_label=extra_bbox_label,
            )
        else:
            self.bboxes_data_matchings = bboxes_data_matchings

    def _get_bboxes_data_matchings(
        self, true_image_data: ImageData, pred_image_data: ImageData, minimum_iou: float, extra_bbox_label: str
    ) -> List[BboxDataMatching]:
        true_bboxes_data = true_image_data.bboxes_data
        pred_bboxes_data = pred_image_data.bboxes_data
        remained_pred_bboxes_data = set(range(len(pred_bboxes_data)))
        bboxes_data_matchings = []

        for tag, bboxes_data in [("true", true_bboxes_data), ("pred", pred_bboxes_data)]:
            bboxes_coords = set()
            for bbox_data in bboxes_data:
                assert all(x is not None for x in [bbox_data.xmin, bbox_data.ymin, bbox_data.xmax, bbox_data.ymax])
                if not (bbox_data.xmin <= bbox_data.xmax and bbox_data.ymin <= bbox_data.ymax):
                    # logger.warning(f"Wrong coordinates {bbox_data.coords=} detected, skipping...")
                    continue
                if bbox_data.coords in bboxes_coords:
                    pass
                    # logger.warning(
                    #     f"Repeated {tag} BboxData with these coords "
                    #     f"(xmin, ymin, xmax, ymax): {bbox_data.coords}. "
                    #     "All BboxData should contain unique elements."
                    # )
                bboxes_coords.add(bbox_data.coords)

        if len(true_bboxes_data) > 0 and len(bboxes_data) > 0:
            pairwise_ious = pairwise_intersection_over_union(true_bboxes_data, bboxes_data)

            for idx, true_bbox_data in enumerate(true_bboxes_data):
                best_pred_bbox_column = np.argmax(pairwise_ious[idx, :])

                if pairwise_ious[idx, best_pred_bbox_column] >= minimum_iou:
                    # Remove pred_bbox_data from pairwise matrix
                    pairwise_ious[:, best_pred_bbox_column] = -1
                    remained_pred_bboxes_data.remove(best_pred_bbox_column)
                else:
                    best_pred_bbox_column = None

                if best_pred_bbox_column is not None:  # Not found
                    bboxes_data_matchings.append(
                        BboxDataMatching(
                            true_bbox_data=true_bbox_data,
                            pred_bbox_data=pred_bboxes_data[best_pred_bbox_column],
                            extra_bbox_label=extra_bbox_label,
                        )
                    )
                else:
                    bboxes_data_matchings.append(
                        BboxDataMatching(  # Found
                            true_bbox_data=true_bbox_data, pred_bbox_data=None, extra_bbox_label=extra_bbox_label
                        )
                    )
            for pred_bbox_data_column in remained_pred_bboxes_data:
                bboxes_data_matchings.append(
                    BboxDataMatching(
                        true_bbox_data=None,
                        pred_bbox_data=pred_bboxes_data[pred_bbox_data_column],
                        extra_bbox_label=extra_bbox_label,
                    )
                )
        elif len(true_bboxes_data) > 0:
            for true_bbox_data in true_bboxes_data:
                bboxes_data_matchings.append(
                    BboxDataMatching(
                        true_bbox_data=true_bbox_data, pred_bbox_data=None, extra_bbox_label=extra_bbox_label
                    )
                )
        elif len(bboxes_data) > 0:
            for pred_bbox_data in bboxes_data:
                bboxes_data_matchings.append(
                    BboxDataMatching(
                        true_bbox_data=None, pred_bbox_data=pred_bbox_data, extra_bbox_label=extra_bbox_label
                    )
                )

        return bboxes_data_matchings

    def get_detection_errors_types(self, label: str = None) -> List[Literal["TP", "FP", "FN"]]:
        detection_errors_types = [
            bbox_data_matching.get_detection_error_type(label=label)
            for bbox_data_matching in self.bboxes_data_matchings
        ]
        detection_errors_types = [error_type for error_type in detection_errors_types if error_type is not None]
        return detection_errors_types

    def get_pipeline_errors_types(
        self, label: str = None
    ) -> List[Literal["TP", "FP", "FN", "TP (extra bbox)", "FP (extra bbox)"]]:
        pipeline_errors_types = [
            bbox_data_matching.get_pipeline_error_type(label=label) for bbox_data_matching in self.bboxes_data_matchings
        ]
        pipeline_errors_types = [error_type for error_type in pipeline_errors_types if error_type is not None]
        return pipeline_errors_types

    def get_detection_TP(self, label: str = None) -> int:
        return self.get_detection_errors_types(label).count("TP")

    def get_detection_FP(self) -> int:
        return self.get_detection_errors_types().count("FP")

    def get_detection_FN(self, label: str = None) -> int:
        return self.get_detection_errors_types(label).count("FN")

    def get_pipeline_TP(self, label: str = None) -> int:
        return self.get_pipeline_errors_types(label).count("TP")

    def get_pipeline_FP(self, label: str = None) -> int:
        return self.get_pipeline_errors_types(label).count("FP")

    def get_pipeline_FN(self, label: str = None) -> int:
        return self.get_pipeline_errors_types(label).count("FN")

    def get_pipeline_TP_extra_bbox(self, label: str = None) -> int:
        return self.get_pipeline_errors_types(label).count("TP (extra bbox)")

    def get_pipeline_FP_extra_bbox(self, label: str = None) -> int:
        return self.get_pipeline_errors_types(label).count("FP (extra bbox)")

    def get_pipeline_FN_extra_bbox(self, label: str = None) -> int:
        return self.get_pipeline_errors_types(label).count("FN (extra bbox)")

    def find_bbox_data_matching(self, bbox_data: BboxData, tag: Literal["true", "pred"]) -> BboxDataMatching:
        xmin, ymin, xmax, ymax = bbox_data.xmin, bbox_data.ymin, bbox_data.xmax, bbox_data.ymax
        if tag == "true":
            bboxes_data_coords_from_matchings = [
                (
                    (
                        bbox_data_matching.true_bbox_data.xmin,
                        bbox_data_matching.true_bbox_data.ymin,
                        bbox_data_matching.true_bbox_data.xmax,
                        bbox_data_matching.true_bbox_data.ymax,
                    )
                    if bbox_data_matching.true_bbox_data is not None
                    else (-1, -1, -1, 1)
                )
                for bbox_data_matching in self.bboxes_data_matchings
            ]
        elif tag == "pred":
            bboxes_data_coords_from_matchings = [
                (
                    (
                        bbox_data_matching.pred_bbox_data.xmin,
                        bbox_data_matching.pred_bbox_data.ymin,
                        bbox_data_matching.pred_bbox_data.xmax,
                        bbox_data_matching.pred_bbox_data.ymax,
                    )
                    if bbox_data_matching.pred_bbox_data is not None
                    else (-1, -1, -1, 1)
                )
                for bbox_data_matching in self.bboxes_data_matchings
            ]
        bbox_data_matching_index = bboxes_data_coords_from_matchings.index((xmin, ymin, xmax, ymax))

        return self.bboxes_data_matchings[bbox_data_matching_index]



import copy
import logging
import os
import sys
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Type, Union

import numpy as np

from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.metrics.image_data_matching import ImageDataMatching
from cv_pipeliner.utils.images_datas import flatten_additional_bboxes_data_in_image_data

# from threading import Semaphore


logger = logging.getLogger("cv_pipeliner.utils.fiftyone")


class FifyOneSession:
    _counter = 0

    def __init__(
        self,
        database_dir: Optional[Union[str, Path]] = None,
        database_uri: Optional[str] = None,
        database_name: Optional[str] = None,
    ):
        assert (
            FifyOneSession._counter == 0
        ), "There is another instance of class FifyOneConverter. Delete it for using this class."
        if "fiftyone" in sys.modules:
            logger.warning(
                "Fiftyone is already imported to some base. The FifyOneConverter will reimport fiftyone and it can change the base."
            )
            del sys.modules["fiftyone"]
        self.database_dir = database_dir
        self.database_uri = database_uri
        self.database_name = database_name
        if self.database_dir is not None:
            os.environ["FIFTYONE_DATABASE_DIR"] = str(self.database_dir)
        if self.database_uri is not None:
            os.environ["FIFTYONE_DATABASE_URI"] = str(self.database_uri)
        if self.database_name is not None:
            os.environ["FIFTYONE_DATABASE_NAME"] = str(self.database_name)

        try:
            import fiftyone

            self._fiftyone = fiftyone
        except Exception as e:
            logger.warning(f"Couldn't connect to fiftyone: {e=}")
            self._fiftyone = None

        FifyOneSession._counter += 1

    @property
    def fiftyone(self):
        if self._fiftyone is None:
            import fiftyone

            self._fiftyone = fiftyone
        return self._fiftyone

    def __del__(self):
        if self.database_dir is not None:
            os.environ.pop("FIFTYONE_DATABASE_DIR", None)
        if self.database_uri is not None:
            os.environ.pop("FIFTYONE_DATABASE_URI", None)
        if self.database_name is not None:
            os.environ.pop("FIFTYONE_DATABASE_NAME", None)
        if "fiftyone" in sys.modules:
            del sys.modules["fiftyone"]
        FifyOneSession._counter -= 1

    def convert_bbox_data_to_fo_detection(
        self,
        bbox_data: BboxData,
        additional_info_keys: List[str] = [],
    ) -> "fiftyone.Detection":
        xminn, yminn, xmaxn, ymaxn = bbox_data.coords_n
        bounding_box = [xminn, yminn, xmaxn - xminn, ymaxn - yminn]
        additional_info = {key: bbox_data.additional_info.get(key, None) for key in additional_info_keys}
        return self.fiftyone.Detection(label=bbox_data.label, bounding_box=bounding_box, **additional_info)

    def convert_bbox_data_keypoints_to_fo_keypoint(self, bbox_data: BboxData) -> "fiftyone.Keypoint":
        if len(bbox_data.keypoints) > 0:
            return self.fiftyone.Keypoint(
                label=bbox_data.label,
                points=[tuple(pair) for pair in bbox_data.keypoints_n],
                source_coords=bbox_data.coords,  # FIXME: https://github.com/voxel51/fiftyone/issues/1610
            )
        else:
            return None

    def convert_image_data_to_fo_detections(
        self,
        image_data: ImageData,
        include_additional_bboxes_data: bool = False,
        additional_info_keys: List[str] = [],
    ) -> "fiftyone.Detections":
        if include_additional_bboxes_data:
            image_data = flatten_additional_bboxes_data_in_image_data(image_data)
        image_data.get_image_size()  # Save to meta
        return self.fiftyone.Detections(
            detections=[
                self.convert_bbox_data_to_fo_detection(bbox_data, additional_info_keys)
                for bbox_data in image_data.bboxes_data
            ]
        )

    def convert_image_data_to_fo_keypoints(
        self, image_data: ImageData, include_additional_bboxes_data: bool = False
    ) -> "fiftyone.Detections":
        if include_additional_bboxes_data:
            image_data = flatten_additional_bboxes_data_in_image_data(image_data)
        image_data.get_image_size()  # Save to meta
        keypoints = (
            [self.fiftyone.Keypoint(label=image_data.label, points=[tuple(pair) for pair in image_data.keypoints_n])]
            if len(image_data.keypoints) > 0
            else []
        ) + [
            fo_keypoints
            for fo_keypoints in map(self.convert_bbox_data_keypoints_to_fo_keypoint, image_data.bboxes_data)
            if fo_keypoints is not None
        ]

        return self.fiftyone.Keypoints(keypoints=keypoints)

    def convert_image_data_to_fo_detections_by_classes(
        self,
        image_data: ImageData,
        include_additional_bboxes_data: bool = False,
    ) -> Dict[str, "fiftyone.Detections"]:
        if include_additional_bboxes_data:
            image_data = flatten_additional_bboxes_data_in_image_data(image_data)
        class_names = sorted(set([bbox_data.label for bbox_data in image_data.bboxes_data]))
        class_name_to_fo_detections = {
            class_name: self.fiftyone.Detections(
                detections=list(
                    map(
                        self.convert_bbox_data_to_fo_detection,
                        [bbox_data for bbox_data in image_data.bboxes_data if bbox_data.label == class_name],
                    )
                )
            )
            for class_name in class_names
        }
        return class_name_to_fo_detections

    def convert_image_data_matching_to_fo_detections(
        self,
        image_data_matching: ImageDataMatching,
        model_type: Literal["detection", "pipeline"],
        label: Optional[str] = None,
    ) -> Tuple["fiftyone.Detections", "fiftyone.Detections"]:
        """
        Convert image_data_matching to FO.Detections, with labels written as "<label> [<error type>]".
        Returns pair of true and pred detections: (true_detections, pred_detections)
        """
        assert model_type in ["detection", "pipeline"]
        image_data_matching = copy.deepcopy(image_data_matching)

        true_detections_instances = []
        pred_detections_instances = []
        for bbox_data_matching in image_data_matching.bboxes_data_matchings:
            if model_type == "detection":
                bbox_error_type = bbox_data_matching.get_detection_error_type(label=label)
            else:
                bbox_error_type = bbox_data_matching.get_pipeline_error_type(label=label)
            true_bbox_data = bbox_data_matching.true_bbox_data
            pred_bbox_data = bbox_data_matching.pred_bbox_data
            if true_bbox_data is not None:
                true_bbox_data.label = f"{true_bbox_data.label} [{bbox_error_type}]"
                true_detections_instances.append(self.convert_bbox_data_to_fo_detection(true_bbox_data))
            if pred_bbox_data is not None:
                pred_bbox_data.label = f"{pred_bbox_data.label} [{bbox_error_type}]"
                pred_detections_instances.append(self.convert_bbox_data_to_fo_detection(pred_bbox_data))

        true_detections = self.fiftyone.Detections(detections=true_detections_instances)
        pred_detections = self.fiftyone.Detections(detections=pred_detections_instances)
        return true_detections, pred_detections

    def convert_image_data_matching_to_fo_detections_by_classes(
        self,
        image_data_matching: ImageDataMatching,
        model_type: Literal["detection", "pipeline"],
        label: Optional[str] = None,
    ) -> Tuple[Dict[str, "fiftyone.Detections"], Dict[str, "fiftyone.Detections"]]:
        """
        Convert image_data_matching to FO.Detections, where bboxes_data_matchings are matched by classes.

        If argument 'label' is set, errors types counts as
        binary classification of [0, 1], where 1 is 'label' and 0 is the other class.

        Returns part of dict [true_dict, pred_dict] of this format:
        {
            'class_type [error_type]': fiftyone.Detections
        }
        """
        assert model_type in ["detection", "pipeline"]
        errors_types = ["TP", "FP", "FN", "TN", "TP (extra bbox)", "FP (extra bbox)"]
        class_names = sorted(
            set(
                [
                    bbox_data_matching.true_bbox_data.label
                    for bbox_data_matching in image_data_matching.bboxes_data_matchings
                    if bbox_data_matching.true_bbox_data is not None
                ]
                + [
                    bbox_data_matching.pred_bbox_data.label
                    for bbox_data_matching in image_data_matching.bboxes_data_matchings
                    if bbox_data_matching.pred_bbox_data is not None
                ]
            )
        )
        class_name_to_fo_true_detections = {
            f"{class_name} [{error_type}]": [] for class_name in class_names for error_type in errors_types
        }
        class_name_to_fo_pred_detections = {
            f"{class_name} [{error_type}]": [] for class_name in class_names for error_type in errors_types
        }
        for bbox_data_matching in image_data_matching.bboxes_data_matchings:
            if model_type == "detection":
                bbox_error_type = bbox_data_matching.get_detection_error_type(label=label)
            else:
                bbox_error_type = bbox_data_matching.get_pipeline_error_type(label=label)

            true_bbox_data = bbox_data_matching.true_bbox_data
            pred_bbox_data = bbox_data_matching.pred_bbox_data
            if true_bbox_data is not None:
                class_name_to_fo_true_detections[f"{true_bbox_data.label} [{bbox_error_type}]"].append(
                    self.convert_bbox_data_to_fo_detection(true_bbox_data)
                )
            if pred_bbox_data is not None:
                class_name_to_fo_pred_detections[f"{pred_bbox_data.label} [{bbox_error_type}]"].append(
                    self.convert_bbox_data_to_fo_detection(pred_bbox_data)
                )

        for key in class_name_to_fo_true_detections:
            class_name_to_fo_true_detections[key] = self.fiftyone.Detections(
                detections=class_name_to_fo_true_detections[key]
            )
        for key in class_name_to_fo_pred_detections:
            class_name_to_fo_pred_detections[key] = self.fiftyone.Detections(
                detections=class_name_to_fo_pred_detections[key]
            )

        return class_name_to_fo_true_detections, class_name_to_fo_pred_detections

    def convert_fo_detection_to_bbox_data(
        self,
        fo_detection: "fiftyone.Detection",
        width: int,
        height: int,
        additional_info_keys_in_fo_detections: List[str] = [],
        bbox_data_cls: Type[BboxData] = BboxData,
    ) -> BboxData:
        xminn, yminn, widthn, heightn = fo_detection.bounding_box
        xmaxn, ymaxn = xminn + widthn, yminn + heightn
        xmin, xmax = xminn * width, xmaxn * width
        ymin, ymax = yminn * height, ymaxn * height
        return bbox_data_cls(
            xmin=xmin,
            ymin=ymin,
            xmax=xmax,
            ymax=ymax,
            meta_width=width,
            meta_height=height,
            label=fo_detection.label,
            additional_info={key: fo_detection[key] for key in additional_info_keys_in_fo_detections},
        )

    def convert_fo_keypoint_to_numpy_keypoints(
        self, fo_keypoint: "fiftyone.Keypoint", width: int, height: int
    ) -> np.array:
        keypoints = np.array(fo_keypoint.points).astype(float).reshape((-1, 2))
        keypoints[:, 0] *= width
        keypoints[:, 1] *= height
        return keypoints

    def convert_image_data_to_fo_sample(
        self,
        image_data: ImageData,
        fo_detections_label: Optional[str] = None,  # берется bboxes_data с их label
        fo_classification_label: Optional[str] = None,  # берется image_data.label
        fo_keypoints_label: Optional[str] = None,  # берется image_data.kepoints + bbox_data.keypoints
        include_additional_bboxes_data: bool = False,
        mapping_filepath: Callable[[str], str] = lambda filepath: filepath,
        additional_info_keys_in_bboxes_data: List[str] = [],
        additional_info_keys_in_image_data: List[str] = [],
        additional_info: Dict[str, Any] = {},
    ) -> "fiftyone.Detections":
        filepath = mapping_filepath(str(image_data.image_path))
        sample = self.fiftyone.Sample(filepath=filepath)
        if fo_detections_label is not None:
            width, height = image_data.get_image_size()  # Save to meta
            sample[fo_detections_label] = self.convert_image_data_to_fo_detections(
                image_data, include_additional_bboxes_data, additional_info_keys_in_bboxes_data
            )
            sample.metadata = self.fiftyone.ImageMetadata(width=width, height=height)
        if fo_classification_label is not None and image_data.label is not None:
            sample[fo_classification_label] = self.fiftyone.Classification(label=image_data.label)
        if fo_keypoints_label is not None:
            sample[fo_keypoints_label] = self.convert_image_data_to_fo_keypoints(
                image_data, include_additional_bboxes_data
            )
        for key in additional_info_keys_in_image_data:
            sample[key] = image_data.additional_info.get(key, None)
        for key, value in additional_info.items():
            sample[key] = value
        return sample

    def convert_sample_to_image_data(
        self,
        sample: "fiftyone.Sample",
        fo_detections_label: Optional[str] = None,
        fo_classification_label: Optional[str] = None,
        fo_keypoints_label: Optional[str] = None,
        mapping_filepath: Callable[[str], str] = lambda filepath: filepath,
        additional_info_keys_in_fo_detections: List[str] = [],
        additional_info_keys_in_sample: List[str] = [],
        image_data_cls: Type[ImageData] = ImageData,
        bbox_data_cls: Type[BboxData] = BboxData,
    ) -> ImageData:
        image_path = mapping_filepath(sample.filepath)
        image_data = image_data_cls(
            image_path=image_path,
            meta_width=sample.metadata.width if sample.metadata else None,
            meta_height=sample.metadata.height if sample.metadata else None,
        )
        if fo_detections_label is not None and (
            sample.has_field(fo_detections_label) and sample[fo_detections_label] is not None
        ):
            width, height = image_data.get_image_size()
            image_data.bboxes_data = [
                self.convert_fo_detection_to_bbox_data(
                    fo_detection, width, height, additional_info_keys_in_fo_detections, bbox_data_cls=bbox_data_cls
                )
                for fo_detection in sample[fo_detections_label].detections
            ]
        if fo_keypoints_label is not None and (
            sample.has_field(fo_keypoints_label) and sample[fo_keypoints_label] is not None
        ):
            coords_to_idx = {bbox_data.coords: idx for idx, bbox_data in enumerate(image_data.bboxes_data)}
            for fo_keypoint in sample[fo_keypoints_label].keypoints:
                keypoints = self.convert_fo_keypoint_to_numpy_keypoints(fo_keypoint, width, height)
                if "source_coords" in fo_keypoint:  # FIXME: https://github.com/voxel51/fiftyone/issues/1610
                    image_data.bboxes_data[coords_to_idx[tuple(fo_keypoint.source_coords)]].keypoints = keypoints
                else:
                    image_data.keypoints = np.append(image_data.keypoints, keypoints)
        if fo_classification_label is not None and (
            sample.has_field(fo_classification_label) and sample[fo_classification_label] is not None
        ):
            image_data.label = sample[fo_classification_label].label
        for key in additional_info_keys_in_sample:
            if sample.has_field(key):
                image_data.additional_info[key] = sample[key]
        return image_data



from typing import List, Literal, Tuple

import matplotlib.pyplot as plt
import numpy as np

from cv_pipeliner.core.data import BboxData
from cv_pipeliner.utils.images import get_img_from_fig


def visualize_bboxes_data(
    bboxes_data: List[BboxData],
    class_name: str = "all",
    visualize_size: int = None,
    pred_bboxes_data: List[BboxData] = None,
    type_only: Literal["TP+FP", "TP", "FP"] = "TP+FP",
    use_random: bool = False,
    draw_rectangle_with_color: Tuple[int, int, int] = None,
) -> np.ndarray:
    true_labels = np.array([bbox_data.label for bbox_data in bboxes_data])
    if pred_bboxes_data is not None:
        assert len(bboxes_data) == len(pred_bboxes_data)
        pred_labels = np.array([bbox_data.label for bbox_data in pred_bboxes_data])

    if class_name != "all":
        data_by_class_name = true_labels == class_name
    else:
        data_by_class_name = np.array([True] * len(bboxes_data))

    if pred_bboxes_data is not None and type_only is not None:
        if type_only == "TP":
            data_by_class_name = data_by_class_name & (true_labels == pred_labels)
        elif type_only == "FP":
            data_by_class_name = data_by_class_name & (true_labels != pred_labels)

    data_by_class_name_idxs = np.where(data_by_class_name)[0]
    if visualize_size is None:
        visualize_size = len(data_by_class_name_idxs)

    size = min(visualize_size, len(data_by_class_name_idxs))

    if use_random:
        idxs = np.random.choice(data_by_class_name_idxs, size=size, replace=False)
    else:
        idxs = data_by_class_name_idxs[:size]

    figsize = 5
    if size >= 10:
        figsize += 20
    fig, axes = plt.subplots(int(np.ceil(size / 5)), 5, figsize=(20, figsize + size // 5))

    for idx, ax in zip(idxs, axes.flatten()):
        bbox_data = bboxes_data[idx]
        bbox = bbox_data.open_cropped_image().copy()
        label = bbox_data.label
        bbox = np.array(bbox)
        ax.imshow(bbox)
        ax.axis("off")
        titles = []
        if label is not None:
            titles.append(f"true: {label}")
        if pred_bboxes_data is not None:
            pred_bbox_data = pred_bboxes_data[idx]
            pred_label = pred_bbox_data.label
            titles.append(f"pred: {pred_label}")
        title = "\n".join(titles)
        ax.set_title(title)

    fig.suptitle(class_name, fontsize=20)

    image = get_img_from_fig(fig)
    plt.close()
    return image



from typing import List

import numpy as np
import pandas as pd


def transpose_columns_and_write_diffs_to_df_with_tags(
    df_with_tags: pd.DataFrame,  # columns are like f'{column_name} [{tag}]'
    columns: List[str],
    tags: List[str],
    compare_tag: str,
    use_colors: bool = True,
):
    assert compare_tag in tags

    transposed_columns = []
    for column in columns:
        for tag in tags:
            transposed_columns.append(f"{column} [{tag}]")

    transposed_df_with_tags = df_with_tags[transposed_columns]

    def round_nan(value, ndigits):
        if not np.isnan(value):
            return round(value, ndigits)
        else:
            return value

    def float_to_int_if_needed(value):
        if isinstance(value, float) and value.is_integer():
            return int(value)
        return value

    for column in columns:
        transposed_df_with_tags.loc[:, f"{column} [{compare_tag}]"] = [
            round_nan(value, 3) for value in transposed_df_with_tags[f"{column} [{compare_tag}]"]
        ]
        compare_tag_values = transposed_df_with_tags[f"{column} [{compare_tag}]"]
        for tag in tags:
            if tag == compare_tag:
                continue
            tag_values = transposed_df_with_tags[f"{column} [{tag}]"]
            tag_diffs = (tag_values.astype(float) - compare_tag_values.astype(float)) * np.reciprocal(
                compare_tag_values.astype(float)
            )

            tag_values = [round_nan(tag_value, 3) for tag_value in tag_values]
            tag_signs = ["+" if tag_diff > 0 else "" for tag_diff in tag_diffs]
            tag_suffixes = [
                (
                    f"({tag_sign}{int(round(100 * tag_diff))}%)"
                    if (not np.isnan(tag_diff) and not np.isinf(tag_diff) and np.abs(tag_diff) > 0.01)
                    else ""
                )
                for tag_sign, tag_diff in zip(tag_signs, tag_diffs)
            ]
            if use_colors:
                tag_colors = ["green" if tag_sign == "+" else "red" for tag_sign in tag_signs]
                tag_suffixes = [
                    f"<font color='{tag_color}'>{tag_suffix}</font>"
                    for tag_color, tag_suffix in zip(tag_colors, tag_suffixes)
                ]

            transposed_df_with_tags.loc[:, f"{column} [{tag}]"] = [
                f"{float_to_int_if_needed(tag_value)} {tag_suffix}"
                for tag_value, tag_suffix in zip(tag_values, tag_suffixes)
            ]

    return transposed_df_with_tags



import json
import tempfile
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Type, Union

import fsspec
import numpy as np
import requests
from joblib import Parallel, delayed
from pathy import Pathy

from cv_pipeliner.core.inference_model import get_preprocess_input_from_script_file
from cv_pipeliner.inference_models.detection.core import (
    DetectionInput,
    DetectionModel,
    DetectionModelSpec,
    DetectionOutput,
)
from cv_pipeliner.utils.files import copy_files_from_directory_to_temp_directory
from cv_pipeliner.utils.images import denormalize_bboxes, get_image_b64


class ObjectDetectionAPI_ModelSpec(DetectionModelSpec):
    config_path: Union[str, Path]
    checkpoint_path: Union[str, Path]
    class_names: Optional[Union[List[str], str, Path]] = None
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    input_size: Union[Tuple[int, int], List[int]] = (None, None)
    device: Optional[str] = None

    @property
    def inference_model_cls(self) -> Type["ObjectDetectionAPI_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.object_detection_api import (
            ObjectDetectionAPI_DetectionModel,
        )

        return ObjectDetectionAPI_DetectionModel


class ObjectDetectionAPI_pb_ModelSpec(DetectionModelSpec):
    saved_model_dir: Union[str, Path]
    input_type: Literal["image_tensor", "float_image_tensor", "encoded_image_string_tensor"]
    class_names: Optional[Union[List[str], str, Path]] = None
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    input_size: Union[Tuple[int, int], List[int]] = (None, None)
    device: Optional[str] = None

    @property
    def inference_model_cls(self) -> Type["ObjectDetectionAPI_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.object_detection_api import (
            ObjectDetectionAPI_DetectionModel,
        )

        return ObjectDetectionAPI_DetectionModel


class ObjectDetectionAPI_TFLite_ModelSpec(DetectionModelSpec):
    model_path: Union[str, Path]
    bboxes_output_index: Union[int, str]
    scores_output_index: Union[int, str]
    classes_output_index: Union[int, str]
    multiclasses_scores_output_index: Optional[Union[int, str]] = None
    class_names: Optional[Union[List[str], str, Path]] = None
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    input_size: Union[Tuple[int, int], List[int]] = (None, None)
    device: Optional[str] = None

    @property
    def inference_model_cls(self) -> Type["ObjectDetectionAPI_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.object_detection_api import (
            ObjectDetectionAPI_DetectionModel,
        )

        return ObjectDetectionAPI_DetectionModel


class ObjectDetectionAPI_KFServing(DetectionModelSpec):
    url: str
    input_name: str
    input_type: Literal["image_tensor", "float_image_tensor", "encoded_image_string_tensor"]
    class_names: Optional[Union[List[str], str, Path]] = None

    @property
    def inference_model_cls(self) -> Type["ObjectDetectionAPI_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.object_detection_api import (
            ObjectDetectionAPI_DetectionModel,
        )

        return ObjectDetectionAPI_DetectionModel


INPUT_TYPE_TO_DTYPE = {
    "image_tensor": np.uint8,
    "float_image_tensor": np.float32,
    "encoded_image_string_tensor": np.uint8,
}


class ObjectDetectionAPI_DetectionModel(DetectionModel):
    def _load_object_detection_api(self, model_spec: ObjectDetectionAPI_ModelSpec):
        import tensorflow as tf
        from object_detection.builders import model_builder
        from object_detection.utils import config_util

        temp_dir = tempfile.TemporaryDirectory()
        temp_dir_path = Path(temp_dir.name)
        model_config_path = temp_dir_path / Pathy.fluid(model_spec.config_path).name
        with open(model_config_path, "wb") as out:
            with fsspec.open(model_spec.config_path, "rb") as src:
                out.write(src.read())
        src_checkpoint_path = Pathy.fluid(model_spec.checkpoint_path)
        checkpoint_path = temp_dir_path / src_checkpoint_path.name
        for src_file in fsspec.open_files(f"{src_checkpoint_path}*", "rb"):
            out_file = temp_dir_path / Pathy.fluid(src_file.path).name
            with open(out_file, "wb") as out:
                with src_file as src:
                    out.write(src.read())
        if model_spec.device is not None:
            self.tf_device = tf.device(model_spec.device)
            self.tf_device.__enter__()
        else:
            self.tf_device = None
        try:
            configs = config_util.get_configs_from_pipeline_file(pipeline_config_path=str(model_config_path))
            model_config = configs["model"]
            self.model = model_builder.build(model_config=model_config, is_training=False)
            ckpt = tf.compat.v2.train.Checkpoint(model=self.model)
            ckpt.restore(str(checkpoint_path)).expect_partial()
        finally:
            if self.tf_device is not None:
                self.tf_device.__exit__()
        self.input_dtype = np.float32

        # Run model through a dummy image so that variables are created
        zeros = np.zeros([640, 640, 3])
        self._raw_predict_single_image_default(zeros)

        temp_dir.cleanup()

    def _load_object_detection_api_pb(self, model_spec: ObjectDetectionAPI_pb_ModelSpec):
        import tensorflow as tf

        temp_folder = copy_files_from_directory_to_temp_directory(directory=model_spec.saved_model_dir)
        temp_folder_path = Path(temp_folder.name)
        if model_spec.device is not None:
            self.tf_device = tf.device(model_spec.device)
            self.tf_device.__enter__()
        else:
            self.tf_device = None
        try:
            self.loaded_model = tf.saved_model.load(str(temp_folder_path))
            self.model = self.loaded_model.signatures["serving_default"]
        finally:
            if self.tf_device is not None:
                self.tf_device.__exit__()
        self.input_dtype = INPUT_TYPE_TO_DTYPE[model_spec.input_type]
        temp_folder.cleanup()

    def _load_object_detection_api_tflite(self, model_spec: ObjectDetectionAPI_TFLite_ModelSpec):
        import tensorflow as tf

        temp_file = tempfile.NamedTemporaryFile()
        with fsspec.open(model_spec.model_path, "rb") as src:
            temp_file.write(src.read())
        model_path = Path(temp_file.name)

        self.model = tf.lite.Interpreter(model_path=str(model_path))
        self.model.allocate_tensors()
        input_detail = self.model.get_input_details()[0]
        self.input_index = input_detail["index"]
        output_details = self.model.get_output_details()
        output_name_to_index = {output_detail["name"]: output_detail["index"] for output_detail in output_details}
        if isinstance(model_spec.bboxes_output_index, str):
            self.bboxes_index = output_name_to_index[model_spec.bboxes_output_index]
        else:
            self.bboxes_index = output_details[model_spec.bboxes_output_index]["index"]
        if isinstance(model_spec.bboxes_output_index, str):
            self.scores_index = output_name_to_index[model_spec.scores_output_index]
        else:
            self.scores_index = output_details[model_spec.scores_output_index]["index"]
        if isinstance(model_spec.classes_output_index, str):
            self.classes_index = output_name_to_index[model_spec.classes_output_index]
        else:
            self.classes_index = output_details[model_spec.classes_output_index]["index"]
        self.input_dtype = input_detail["dtype"]
        temp_file.close()

    def __init__(
        self,
        model_spec: Union[
            ObjectDetectionAPI_ModelSpec,
            ObjectDetectionAPI_pb_ModelSpec,
            ObjectDetectionAPI_TFLite_ModelSpec,
            ObjectDetectionAPI_KFServing,
        ],
    ):
        super().__init__(model_spec)

        if model_spec.class_names is not None:
            if isinstance(model_spec.class_names, str) or isinstance(model_spec.class_names, Path):
                with fsspec.open(model_spec.class_names, "r", encoding="utf-8") as out:
                    self.class_names = np.array(json.load(out))
            else:
                self.class_names = np.array(model_spec.class_names)

            if isinstance(model_spec, ObjectDetectionAPI_pb_ModelSpec):
                self.class_names_coef = -1  # saved_model.pb returns from 1
            else:
                self.class_names_coef = 0
        else:
            self.class_names = None
            self.coef = -1

        if isinstance(model_spec, ObjectDetectionAPI_ModelSpec):
            self._load_object_detection_api(model_spec)
            self._raw_predict_single_image = self._raw_predict_single_image_default
        elif isinstance(model_spec, ObjectDetectionAPI_pb_ModelSpec):
            self._load_object_detection_api_pb(model_spec)
            self._raw_predict_single_image = self._raw_predict_single_image_default
        elif isinstance(model_spec, ObjectDetectionAPI_TFLite_ModelSpec):
            self._load_object_detection_api_tflite(model_spec)
            self._raw_predict_single_image = self._raw_predict_single_image_tflite
        elif isinstance(model_spec, ObjectDetectionAPI_KFServing):
            self.input_dtype = INPUT_TYPE_TO_DTYPE[model_spec.input_type]
            # Wake up the service
            try:
                self._raw_predict_single_image_kfserving(image=np.zeros((128, 128, 3)), timeout=1.0)
            except requests.exceptions.ReadTimeout:
                pass
            self._raw_predict_single_image = self._raw_predict_single_image_kfserving
        else:
            raise ValueError(f"ObjectDetectionAPI_Model got unknown DetectionModelSpec: {type(model_spec)}")

        if isinstance(model_spec.preprocess_input, str) or isinstance(model_spec.preprocess_input, Path):
            self._preprocess_input = get_preprocess_input_from_script_file(script_file=model_spec.preprocess_input)
        else:
            if model_spec.preprocess_input is None:
                self._preprocess_input = lambda x: x
            else:
                self._preprocess_input = model_spec.preprocess_input

    def _parse_detection_output_dict(
        self, detection_output_dict: Dict[str, Any]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        raw_bboxes = detection_output_dict["detection_boxes"][0]  # (ymin, xmin, ymax, xmax)
        raw_bboxes = np.array(raw_bboxes)[:, [1, 0, 3, 2]]  # (xmin, ymin, xmax, ymax)
        raw_keypoints = np.array([]).reshape(len(raw_bboxes), 0, 2)
        raw_scores = detection_output_dict["detection_scores"][0]
        raw_scores = np.array(raw_scores)
        raw_classes = detection_output_dict["detection_classes"][0]
        raw_classes = np.array(raw_classes)
        return raw_bboxes, raw_keypoints, raw_scores, raw_classes

    def _raw_predict_single_image_default(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        import tensorflow as tf

        if self.tf_device is not None:
            self.tf_device.__enter__()
        try:
            input_tensor = tf.convert_to_tensor(image, dtype=self.input_dtype)
            if (
                isinstance(self.model_spec, ObjectDetectionAPI_pb_ModelSpec)
                and self.model_spec.input_type == "encoded_image_string_tensor"
            ):
                input_tensor = tf.io.encode_jpeg(input_tensor, quality=100)
            input_tensor = input_tensor[None, ...]
            detection_output_dict = self.model(input_tensor)
        finally:
            if self.tf_device is not None:
                self.tf_device.__exit__()

        return self._parse_detection_output_dict(detection_output_dict)

    def _raw_predict_single_image_tflite(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        height, width, _ = image.shape
        image = np.array(image[None, ...], dtype=self.input_dtype)
        self.model.resize_tensor_input(0, [1, height, width, 3])
        self.model.allocate_tensors()
        self.model.set_tensor(self.input_index, image)
        self.model.invoke()

        raw_bboxes = np.array(self.model.get_tensor(self.bboxes_index))[0]  # (ymin, xmin, ymax, xmax)
        raw_keypoints = np.array([]).reshape(len(raw_bboxes), 0, 2)
        raw_bboxes = raw_bboxes[:, [1, 0, 3, 2]]  # (xmin, ymin, xmax, ymax)
        raw_scores = np.array(self.model.get_tensor(self.scores_index))[0]
        raw_classes = np.array(self.model.get_tensor(self.classes_index))[0]

        return raw_bboxes, raw_keypoints, raw_scores, raw_classes

    def _raw_predict_single_image_kfserving(
        self, image: np.ndarray, timeout: Union[float, None] = None
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if self.model_spec.input_type in ["float_image_tensor", "image_tensor"]:
            input_data = {"inputs": {self.model_spec.input_name: [np.array(image).astype(self.input_dtype).tolist()]}}
        elif self.model_spec.input_type == "encoded_image_string_tensor":
            input_data = {
                "instances": [{self.model_spec.input_name: {"b64": get_image_b64(image, "JPEG", quality=95)}}]
            }
        response = requests.post(url=self.model_spec.url, json=input_data, timeout=timeout)
        try:
            detection_output_dict = response.json()
        except JSONDecodeError:
            raise ValueError(f"Failed to decode JSON. Response content: {response.content}")
        if not response.ok:
            raise ValueError(f"Response is not ok: {response.status_code=}; {response.content=}")
        if "outputs" in detection_output_dict:
            detection_output_dict = detection_output_dict["outputs"]
        elif "predictions" in detection_output_dict:
            detection_output_dict = detection_output_dict["predictions"]
            detection_output_dict = {
                "detection_boxes": [detection_output_dict[0]["detection_boxes"]],
                "detection_scores": [detection_output_dict[0]["detection_scores"]],
                "detection_classes": [detection_output_dict[0]["detection_classes"]],
            }
        return self._parse_detection_output_dict(detection_output_dict)

    def _postprocess_prediction(
        self,
        raw_bboxes: np.ndarray,
        raw_keypoints: np.ndarray,
        raw_scores: np.ndarray,
        raw_classes: np.ndarray,
        score_threshold: float,
        width: int,
        height: int,
        classification_top_n: int,
    ) -> Tuple[
        List[Tuple[int, int, int, int]], List[List[Tuple[int, int]]], List[float], List[List[str]], List[List[float]]
    ]:
        raw_bboxes = denormalize_bboxes(raw_bboxes, width, height)
        mask = raw_scores > score_threshold
        bboxes = raw_bboxes[mask]
        keypoints = raw_keypoints[mask]
        scores = raw_scores[mask]
        classes = raw_classes[mask]

        correct_non_repeated_bboxes_idxs = []
        bboxes_set = set()
        for idx, bbox in enumerate(bboxes):
            xmin, ymin, xmax, ymax = bbox
            if xmax - xmin > 0 and ymax - ymin > 0 and (xmin, ymin, xmax, ymax) not in bboxes_set:
                bboxes_set.add((xmin, ymin, xmax, ymax))
                correct_non_repeated_bboxes_idxs.append(idx)

        bboxes = bboxes[correct_non_repeated_bboxes_idxs]
        keypoints = keypoints[correct_non_repeated_bboxes_idxs]
        scores = scores[correct_non_repeated_bboxes_idxs]
        classes = classes[correct_non_repeated_bboxes_idxs]
        classes_scores = scores.copy()
        if self.class_names is not None:
            class_names_top_n = np.array(
                [
                    [class_name for i in range(classification_top_n)]
                    for class_name in self.class_names[(classes.astype(np.int32) + self.class_names_coef)]
                ]
            )
            classes_scores_top_n = np.array([[score for _ in range(classification_top_n)] for score in classes_scores])
        else:
            class_names_top_n = np.array([[None for _ in range(classification_top_n)] for _ in classes])
            classes_scores_top_n = np.array([[score for _ in range(classification_top_n)] for score in classes_scores])

        return bboxes, keypoints, scores, class_names_top_n, classes_scores_top_n

    def _predict_single_image(
        self, image: np.ndarray, score_threshold: float, classification_top_n: int = 1
    ) -> DetectionOutput:
        height, width, _ = image.shape
        input_image = self.preprocess_input([image])[0]
        raw_bboxes, raw_keypoints, raw_scores, raw_classes = self._raw_predict_single_image(input_image)
        bboxes, keypoints, scores, class_names_top_k, classes_scores_top_k = self._postprocess_prediction(
            raw_bboxes=raw_bboxes,
            raw_keypoints=raw_keypoints,
            raw_scores=raw_scores,
            raw_classes=raw_classes,
            score_threshold=score_threshold,
            width=width,
            height=height,
            classification_top_n=classification_top_n,
        )
        return bboxes, keypoints, scores, class_names_top_k, classes_scores_top_k

    def predict(
        self, input: DetectionInput, score_threshold: float, classification_top_n: int = 1, n_jobs: int = 1
    ) -> DetectionOutput:
        results = Parallel(n_jobs=n_jobs, prefer="threads")(
            delayed(self._predict_single_image)(
                image=image,
                score_threshold=score_threshold,
                classification_top_n=classification_top_n,
            )
            for image in input
        )
        (n_pred_bboxes, n_pred_keypoints, n_pred_scores, n_pred_class_names_top_k, n_pred_scores_top_k) = (
            [res[i] for res in results] for i in range(5)
        )
        n_pred_masks = [[[] for _ in pred_bboxes] for pred_bboxes in n_pred_bboxes]
        return (
            n_pred_bboxes,
            n_pred_keypoints,
            n_pred_masks,
            n_pred_scores,
            n_pred_class_names_top_k,
            n_pred_scores_top_k,
        )

    def preprocess_input(self, input: DetectionInput):
        return self._preprocess_input(input)

    @property
    def input_size(self) -> Tuple[int, int]:
        return self.model_spec.input_size



import tempfile
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Callable, List, Tuple, Union

import imageio
import imutils
import numpy as np
from tqdm import tqdm

from cv_pipeliner.batch_generators.bbox_data import BatchGeneratorBboxData
from cv_pipeliner.batch_generators.image_data import BatchGeneratorImageData
from cv_pipeliner.core.data import BboxData, ImageData
from cv_pipeliner.inferencers.classification import ClassificationInferencer
from cv_pipeliner.inferencers.detection import DetectionInferencer
from cv_pipeliner.inferencers.pipeline import PipelineInferencer
from cv_pipeliner.tracking.opencv_tracker import OpenCVTracker
from cv_pipeliner.tracking.sort_tracker import Sort
from cv_pipeliner.visualizers.core.image_data import visualize_image_data


@dataclass
class FrameResult:
    label: str
    track_id: int
    ready_at_frame: int


class VideoInferencer:
    def __init__(
        self,
        pipeline_inferencer: PipelineInferencer,
        draw_base_labels_with_given_label_to_base_label_image: Callable[[str], np.ndarray],
        write_labels: bool = True,
        frame_width: int = 640,
        frame_height: int = 1152,
        batch_size: int = 16,
    ):
        self.detection_inferencer = DetectionInferencer(pipeline_inferencer.model.detection_model)
        self.classification_inferencer = ClassificationInferencer(pipeline_inferencer.model.classification_model)
        self.draw_base_labels_with_given_label_to_base_label_image = (
            draw_base_labels_with_given_label_to_base_label_image
        )
        self.write_labels = write_labels

        self.sort_tracker = None
        self.opencv_tracker = None

        self.frame_width = frame_width
        self.frame_height = frame_height

        self.current_ready_frames_queue: List[FrameResult] = []

    def update_sort_tracker(
        self, bboxes: List[Tuple[int, int, int, int]], scores: List[float] = None
    ) -> Tuple[List[Tuple[int, int, int, int]], List[int]]:
        if scores is None:
            # SORT requires detection scores, use 1 while tracking
            scores = np.ones(bboxes.shape[0])
        tracked_bboxes_sort = self.sort_tracker.update(np.column_stack([bboxes, scores]))
        tracked_bboxes_sort = tracked_bboxes_sort.round().astype(int)
        tracked_bboxes = tracked_bboxes_sort[:, :-1]
        tracked_bboxes = np.clip(tracked_bboxes, 0, tracked_bboxes.max(initial=0))
        tracked_ids = tracked_bboxes_sort[:, -1]
        return tracked_bboxes, tracked_ids

    def run_tracking_on_frame(self, frame: np.ndarray) -> Tuple[List[Tuple[int, int, int, int]], List[int]]:
        tracked_bboxes_optical = self.opencv_tracker.update(frame)
        tracked_bboxes, tracked_ids = self.update_sort_tracker(bboxes=tracked_bboxes_optical)
        return tracked_bboxes, tracked_ids

    def run_pipeline_on_frame(
        self,
        frame: np.ndarray,
        frame_idx: int,
        fps: float,
        detection_delay: int,
        classification_delay: int,
        detection_score_threshold: float,
        batch_size: int,
    ) -> Tuple[List[Tuple[int, int, int, int]], List[int]]:
        frame = frame.copy()
        image_data = ImageData(image=frame)
        image_data_gen = BatchGeneratorImageData(
            [image_data], batch_size=batch_size, use_not_caught_elements_as_last_batch=True
        )

        pred_image_data = self.detection_inferencer.predict(
            images_data_gen=image_data_gen, score_threshold=detection_score_threshold
        )[0]
        bboxes = np.array(
            [
                (bbox_data.xmin, bbox_data.ymin, bbox_data.xmax, bbox_data.ymax)
                for bbox_data in pred_image_data.bboxes_data
            ]
        )
        detection_scores = np.array([bbox_data.detection_score for bbox_data in pred_image_data.bboxes_data])

        self.opencv_tracker = OpenCVTracker(bboxes, frame)
        tracked_bboxes, tracked_ids = self.update_sort_tracker(bboxes=bboxes, scores=detection_scores)

        # detection_delay_frames = int(round(detection_delay * fps / 1000))
        # classification_delay_frames = int(round(classification_delay * fps / 1000))

        current_tracks_ids = [frame_result.track_id for frame_result in self.current_ready_frames_queue]
        current_not_tracked_items_idxs = [
            idx for idx, tracked_id in enumerate(tracked_ids) if tracked_id not in current_tracks_ids
        ]
        if current_not_tracked_items_idxs:
            current_not_tracked_bboxes = tracked_bboxes[current_not_tracked_items_idxs]
            current_not_tracked_ids = tracked_ids[current_not_tracked_items_idxs]
            bboxes_data = [
                BboxData(image=frame, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
                for (xmin, ymin, xmax, ymax) in current_not_tracked_bboxes
            ]
            bboxes_data_gen = BatchGeneratorBboxData(
                [bboxes_data], batch_size=batch_size, use_not_caught_elements_as_last_batch=True
            )
            pred_bboxes_data = self.classification_inferencer.predict(bboxes_data_gen)[0]

            for bbox_data, tracked_id in zip(pred_bboxes_data, current_not_tracked_ids):
                ready_at_frame = frame_idx
                frame_result = FrameResult(label=bbox_data.label, track_id=tracked_id, ready_at_frame=ready_at_frame)
                self.current_ready_frames_queue.append(frame_result)

        return tracked_bboxes, tracked_ids

    def draw_overlay(
        self,
        frame: np.ndarray,
        tracked_bboxes: List[Tuple[int, int, int, int]],
        tracked_ids: List[int],
        ready_frames_at_the_moment: List["FrameResult"],
        filter_by_labels: List[str] = None,
    ) -> np.ndarray:
        image = frame.copy()
        tracked_bboxes = tracked_bboxes.astype(int)
        ready_tracks_ids_at_the_moment = [ready_frame.track_id for ready_frame in ready_frames_at_the_moment]

        current_bboxes_data = []
        for bbox, track_id in zip(tracked_bboxes, tracked_ids):
            if track_id not in ready_tracks_ids_at_the_moment:
                continue

            xmin, ymin, xmax, ymax = bbox
            ready_frame = ready_frames_at_the_moment[ready_tracks_ids_at_the_moment.index(track_id)]
            label = ready_frame.label
            current_bbox_data = BboxData(image=image, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, label=label)
            current_bboxes_data.append(current_bbox_data)

        image_data = ImageData(image=frame, bboxes_data=current_bboxes_data)

        image = visualize_image_data(
            image_data=image_data,
            include_labels=self.write_labels,
            filter_by_labels=filter_by_labels,
            draw_base_labels_with_given_label_to_base_label_image=(
                self.draw_base_labels_with_given_label_to_base_label_image
            ),
            known_labels=self.classification_inferencer.class_names,
        )

        return image

    def process_video(
        self,
        video_file: Union[str, Path, BytesIO],
        classification_delay: int,
        detection_delay: int,
        detection_score_threshold: float,
        filter_by_labels: List[str],
        disable_tqdm: bool = False,
        batch_size: int = 16,
    ) -> tempfile.NamedTemporaryFile:
        result = []

        self.sort_tracker = Sort()
        with imageio.get_reader(video_file, ".mp4") as reader:
            fps = reader.get_meta_data()["fps"]

            for frame_idx, frame in tqdm(list(enumerate(reader)), disable=disable_tqdm):
                frame = imutils.resize(frame, width=self.frame_height, height=self.frame_height)

                detection_delay_frames = detection_delay * fps / 1000
                if frame_idx % detection_delay_frames == 0:
                    tracked_bboxes, tracked_ids = self.run_pipeline_on_frame(
                        frame=frame,
                        frame_idx=frame_idx,
                        fps=fps,
                        detection_delay=detection_delay,
                        classification_delay=classification_delay,
                        detection_score_threshold=detection_score_threshold,
                        batch_size=batch_size,
                    )
                else:
                    tracked_bboxes, tracked_ids = self.run_tracking_on_frame(frame)

                ready_frames_at_the_moment = [
                    ready_frame
                    for ready_frame in self.current_ready_frames_queue
                    if ready_frame.ready_at_frame <= frame_idx
                ]

                result_frame = self.draw_overlay(
                    frame=frame,
                    tracked_bboxes=tracked_bboxes,
                    tracked_ids=tracked_ids,
                    ready_frames_at_the_moment=ready_frames_at_the_moment,
                    filter_by_labels=filter_by_labels,
                )
                result.append(result_frame)

        temp_file = tempfile.NamedTemporaryFile(suffix=".mp4").name
        imageio.mimwrite(temp_file, result, codec="h264", fps=fps)
        return temp_file



from pathlib import Path
from typing import Callable, List, Tuple, Type, Union

import numpy as np
from pathy import Pathy

from cv_pipeliner.core.inference_model import get_preprocess_input_from_script_file
from cv_pipeliner.inference_models.embedder.core import (
    EmbedderInput,
    EmbedderModel,
    EmbedderModelSpec,
    EmbedderOutput,
)


class PyTorch_EmbedderModelSpec(EmbedderModelSpec):
    model_path: Union[str, Pathy]
    device: str
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None

    @property
    def inference_model_cls(self) -> Type["PyTorch_EmbedderModel"]:
        from cv_pipeliner.inference_models.embedder.pytorch import PyTorch_EmbedderModel

        return PyTorch_EmbedderModel


class PyTorch_EmbedderModel(EmbedderModel):
    def _load_pytorch_model_spec(self, model_spec: PyTorch_EmbedderModelSpec):
        import torch

        self.device = torch.device(model_spec.device)
        self.model = torch.jit.load(str(model_spec.model_path)).to(self.device)
        self.model.eval()

    def __init__(self, model_spec: PyTorch_EmbedderModelSpec):
        super().__init__(model_spec)

        self._load_pytorch_model_spec(model_spec)
        self._raw_predict = self._raw_predict_pytorch

        if isinstance(model_spec.preprocess_input, str) or isinstance(model_spec.preprocess_input, Path):
            self._preprocess_input = get_preprocess_input_from_script_file(script_file=model_spec.preprocess_input)
        else:
            if model_spec.preprocess_input is None:
                self._preprocess_input = lambda x: x
            else:
                self._preprocess_input = model_spec.preprocess_input

    def _raw_predict_pytorch(self, images: np.ndarray):
        import torch

        outputs = []
        with torch.no_grad():
            images = images.float().to(self.device)
            prediction = self.model(images).cpu().detach().numpy()
            outputs.extend(prediction)
        return outputs

    def preprocess_input(self, input: EmbedderInput):
        return self._preprocess_input(input)

    def predict(
        self,
        input: EmbedderInput,
    ) -> EmbedderOutput:
        input = self.preprocess_input(input)
        predictions = self._raw_predict_pytorch(input)
        return predictions

    @property
    def input_size(self) -> Tuple[int, int]:
        return self.model_spec.input_size



import abc
from typing import Any, Union

from cv_pipeliner.core.batch_generator import BatchGenerator
from cv_pipeliner.core.inference_model import InferenceModel


class Inferencer(abc.ABC):
    """
    High-level class for inference.

    The class accepts loaded InferenceModel and make predictions
    by using high-level objects (BatchGenerators).

    Example:
        model_spec = ModelSpec(...)
        inference_model = model_spec.load_model()
        inferencer = Inferencer(inference_model)
        data_gen = BatchGenerator(data, batch_size=16)
        pred_data = inferencer.predict(data_gen)  # or inferencer.predict(data), if possible

    """

    def __init__(self, model: InferenceModel):
        self.model = model

    @abc.abstractmethod
    def predict(self, data: Union[Any, BatchGenerator]):
        pass



import math
import warnings

import cv2
import numpy as np

# Taken from https://github.com/open-mmlab/mmpose/blob/master/mmpose/core/evaluation/top_down_eval.py


def _gaussian_blur(heatmaps, kernel=11):
    """Modulate heatmap distribution with Gaussian.
     sigma = 0.3*((kernel_size-1)*0.5-1)+0.8
     sigma~=3 if k=17
     sigma=2 if k=11;
     sigma~=1.5 if k=7;
     sigma~=1 if k=3;
    Note:
        - batch_size: N
        - num_keypoints: K
        - heatmap height: H
        - heatmap width: W
    Args:
        heatmaps (np.ndarray[N, K, H, W]): model predicted heatmaps.
        kernel (int): Gaussian kernel size (K) for modulation, which should
            match the heatmap gaussian sigma when training.
            K=17 for sigma=3 and k=11 for sigma=2.
    Returns:
        np.ndarray ([N, K, H, W]): Modulated heatmap distribution.
    """
    assert kernel % 2 == 1

    border = (kernel - 1) // 2
    batch_size = heatmaps.shape[0]
    num_joints = heatmaps.shape[1]
    height = heatmaps.shape[2]
    width = heatmaps.shape[3]
    for i in range(batch_size):
        for j in range(num_joints):
            origin_max = np.max(heatmaps[i, j])
            dr = np.zeros((height + 2 * border, width + 2 * border), dtype=np.float32)
            dr[border:-border, border:-border] = heatmaps[i, j].copy()
            dr = cv2.GaussianBlur(dr, (kernel, kernel), 0)
            heatmaps[i, j] = dr[border:-border, border:-border].copy()
            heatmaps[i, j] *= origin_max / np.max(heatmaps[i, j])
    return heatmaps


def _taylor(heatmap, coord):
    """Distribution aware coordinate decoding method.
    Note:
        - heatmap height: H
        - heatmap width: W
    Args:
        heatmap (np.ndarray[H, W]): Heatmap of a particular joint type.
        coord (np.ndarray[2,]): Coordinates of the predicted keypoints.
    Returns:
        np.ndarray[2,]: Updated coordinates.
    """
    H, W = heatmap.shape[:2]
    px, py = int(coord[0]), int(coord[1])
    if 1 < px < W - 2 and 1 < py < H - 2:
        dx = 0.5 * (heatmap[py][px + 1] - heatmap[py][px - 1])
        dy = 0.5 * (heatmap[py + 1][px] - heatmap[py - 1][px])
        dxx = 0.25 * (heatmap[py][px + 2] - 2 * heatmap[py][px] + heatmap[py][px - 2])
        dxy = 0.25 * (
            heatmap[py + 1][px + 1] - heatmap[py - 1][px + 1] - heatmap[py + 1][px - 1] + heatmap[py - 1][px - 1]
        )
        dyy = 0.25 * (heatmap[py + 2 * 1][px] - 2 * heatmap[py][px] + heatmap[py - 2 * 1][px])
        derivative = np.array([[dx], [dy]])
        hessian = np.array([[dxx, dxy], [dxy, dyy]])
        if dxx * dyy - dxy**2 != 0:
            hessianinv = np.linalg.inv(hessian)
            offset = -hessianinv @ derivative
            offset = np.squeeze(np.array(offset.T), axis=0)
            coord += offset
    return coord


def post_dark_udp(coords, batch_heatmaps, kernel=3):
    """DARK post-pocessing. Implemented by udp. Paper ref: Huang et al. The
    Devil is in the Details: Delving into Unbiased Data Processing for Human
    Pose Estimation (CVPR 2020). Zhang et al. Distribution-Aware Coordinate
    Representation for Human Pose Estimation (CVPR 2020).
    Note:
        - batch size: B
        - num keypoints: K
        - num persons: N
        - height of heatmaps: H
        - width of heatmaps: W
        B=1 for bottom_up paradigm where all persons share the same heatmap.
        B=N for top_down paradigm where each person has its own heatmaps.
    Args:
        coords (np.ndarray[N, K, 2]): Initial coordinates of human pose.
        batch_heatmaps (np.ndarray[B, K, H, W]): batch_heatmaps
        kernel (int): Gaussian kernel size (K) for modulation.
    Returns:
        np.ndarray([N, K, 2]): Refined coordinates.
    """
    if not isinstance(batch_heatmaps, np.ndarray):
        batch_heatmaps = batch_heatmaps.cpu().numpy()
    B, K, H, W = batch_heatmaps.shape
    N = coords.shape[0]
    assert B == 1 or B == N
    for heatmaps in batch_heatmaps:
        for heatmap in heatmaps:
            cv2.GaussianBlur(heatmap, (kernel, kernel), 0, heatmap)
    np.clip(batch_heatmaps, 0.001, 50, batch_heatmaps)
    np.log(batch_heatmaps, batch_heatmaps)

    batch_heatmaps_pad = np.pad(batch_heatmaps, ((0, 0), (0, 0), (1, 1), (1, 1)), mode="edge").flatten()

    index = coords[..., 0] + 1 + (coords[..., 1] + 1) * (W + 2)
    index += (W + 2) * (H + 2) * np.arange(0, B * K).reshape(-1, K)
    index = index.astype(int).reshape(-1, 1)
    i_ = batch_heatmaps_pad[index]
    ix1 = batch_heatmaps_pad[index + 1]
    iy1 = batch_heatmaps_pad[index + W + 2]
    ix1y1 = batch_heatmaps_pad[index + W + 3]
    ix1_y1_ = batch_heatmaps_pad[index - W - 3]
    ix1_ = batch_heatmaps_pad[index - 1]
    iy1_ = batch_heatmaps_pad[index - 2 - W]

    dx = 0.5 * (ix1 - ix1_)
    dy = 0.5 * (iy1 - iy1_)
    derivative = np.concatenate([dx, dy], axis=1)
    derivative = derivative.reshape(N, K, 2, 1)
    dxx = ix1 - 2 * i_ + ix1_
    dyy = iy1 - 2 * i_ + iy1_
    dxy = 0.5 * (ix1y1 - ix1 - iy1 + i_ + i_ - ix1_ - iy1_ + ix1_y1_)
    hessian = np.concatenate([dxx, dxy, dxy, dyy], axis=1)
    hessian = hessian.reshape(N, K, 2, 2)
    hessian = np.linalg.inv(hessian + np.finfo(np.float32).eps * np.eye(2))
    coords -= np.einsum("ijmn,ijnk->ijmk", hessian, derivative).squeeze()
    return coords


def bbox_xywh2cs(bbox, aspect_ratio, padding=1.0, pixel_std=200.0):
    """Transform the bbox format from (x,y,w,h) into (center, scale)
    Args:
        bbox (ndarray): Single bbox in (x, y, w, h)
        aspect_ratio (float): The expected bbox aspect ratio (w over h)
        padding (float): Bbox padding factor that will be multilied to scale.
            Default: 1.0
        pixel_std (float): The scale normalization factor. Default: 200.0
    Returns:
        tuple: A tuple containing center and scale.
        - np.ndarray[float32](2,): Center of the bbox (x, y).
        - np.ndarray[float32](2,): Scale of the bbox w & h.
    """

    x, y, w, h = bbox[:4]
    center = np.array([x + w * 0.5, y + h * 0.5], dtype=np.float32)

    if w > aspect_ratio * h:
        h = w * 1.0 / aspect_ratio
    elif w < aspect_ratio * h:
        w = h * aspect_ratio

    scale = np.array([w, h], dtype=np.float32) / pixel_std
    scale = scale * padding

    return center, scale


def _get_3rd_point(a, b):
    """To calculate the affine matrix, three pairs of points are required. This
    function is used to get the 3rd point, given 2D points a & b.

    The 3rd point is defined by rotating vector `a - b` by 90 degrees
    anticlockwise, using b as the rotation center.

    Args:
        a (np.ndarray): point(x,y)
        b (np.ndarray): point(x,y)

    Returns:
        np.ndarray: The 3rd point.
    """
    assert len(a) == 2
    assert len(b) == 2
    direction = a - b
    third_pt = b + np.array([-direction[1], direction[0]], dtype=np.float32)

    return third_pt


def rotate_point(pt, angle_rad):
    """Rotate a point by an angle.

    Args:
        pt (list[float]): 2 dimensional point to be rotated
        angle_rad (float): rotation angle by radian

    Returns:
        list[float]: Rotated point.
    """
    assert len(pt) == 2
    sn, cs = np.sin(angle_rad), np.cos(angle_rad)
    new_x = pt[0] * cs - pt[1] * sn
    new_y = pt[0] * sn + pt[1] * cs
    rotated_pt = [new_x, new_y]

    return rotated_pt


def get_affine_transform(center, scale, rot, output_size, shift=(0.0, 0.0), inv=False):
    """Get the affine transform matrix, given the center/scale/rot/output_size.

    Args:
        center (np.ndarray[2, ]): Center of the bounding box (x, y).
        scale (np.ndarray[2, ]): Scale of the bounding box
            wrt [width, height].
        rot (float): Rotation angle (degree).
        output_size (np.ndarray[2, ] | list(2,)): Size of the
            destination heatmaps.
        shift (0-100%): Shift translation ratio wrt the width/height.
            Default (0., 0.).
        inv (bool): Option to inverse the affine transform direction.
            (inv=False: src->dst or inv=True: dst->src)

    Returns:
        np.ndarray: The transform matrix.
    """
    assert len(center) == 2
    assert len(scale) == 2
    assert len(output_size) == 2
    assert len(shift) == 2

    # pixel_std is 200.
    scale_tmp = scale * 200.0

    shift = np.array(shift)
    src_w = scale_tmp[0]
    dst_w = output_size[0]
    dst_h = output_size[1]

    rot_rad = np.pi * rot / 180
    src_dir = rotate_point([0.0, src_w * -0.5], rot_rad)
    dst_dir = np.array([0.0, dst_w * -0.5])

    src = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale_tmp * shift
    src[1, :] = center + src_dir + scale_tmp * shift
    src[2, :] = _get_3rd_point(src[0, :], src[1, :])

    dst = np.zeros((3, 2), dtype=np.float32)
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir
    dst[2, :] = _get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    return trans


def get_warp_matrix(theta, size_input, size_dst, size_target):
    """Calculate the transformation matrix under the constraint of unbiased.
    Paper ref: Huang et al. The Devil is in the Details: Delving into Unbiased
    Data Processing for Human Pose Estimation (CVPR 2020).

    Args:
        theta (float): Rotation angle in degrees.
        size_input (np.ndarray): Size of input image [w, h].
        size_dst (np.ndarray): Size of output image [w, h].
        size_target (np.ndarray): Size of ROI in input plane [w, h].

    Returns:
        np.ndarray: A matrix for transformation.
    """
    theta = np.deg2rad(theta)
    matrix = np.zeros((2, 3), dtype=np.float32)
    scale_x = size_dst[0] / size_target[0]
    scale_y = size_dst[1] / size_target[1]
    matrix[0, 0] = math.cos(theta) * scale_x
    matrix[0, 1] = -math.sin(theta) * scale_x
    matrix[0, 2] = scale_x * (
        -0.5 * size_input[0] * math.cos(theta) + 0.5 * size_input[1] * math.sin(theta) + 0.5 * size_target[0]
    )
    matrix[1, 0] = math.sin(theta) * scale_y
    matrix[1, 1] = math.cos(theta) * scale_y
    matrix[1, 2] = scale_y * (
        -0.5 * size_input[0] * math.sin(theta) - 0.5 * size_input[1] * math.cos(theta) + 0.5 * size_target[1]
    )
    return matrix


def top_down_affine(img, r, c, s, image_size, use_udp):
    if use_udp:
        trans = get_warp_matrix(r, np.array(c) * 2.0, np.array(image_size) - 1.0, np.array(s) * 200.0)
        if not isinstance(img, list):
            img = cv2.warpAffine(img, trans, (int(image_size[0]), int(image_size[1])), flags=cv2.INTER_LINEAR)
        else:
            img = [
                cv2.warpAffine(i, trans, (int(image_size[0]), int(image_size[1])), flags=cv2.INTER_LINEAR) for i in img
            ]

    else:
        trans = get_affine_transform(c, s, r, image_size)
        if not isinstance(img, list):
            img = cv2.warpAffine(img, trans, (int(image_size[0]), int(image_size[1])), flags=cv2.INTER_LINEAR)
        else:
            img = [
                cv2.warpAffine(i, trans, (int(image_size[0]), int(image_size[1])), flags=cv2.INTER_LINEAR) for i in img
            ]

    return img


def _get_max_preds(heatmaps):
    """Get keypoint predictions from score maps.
    Note:
        batch_size: N
        num_keypoints: K
        heatmap height: H
        heatmap width: W
    Args:
        heatmaps (np.ndarray[N, K, H, W]): model predicted heatmaps.
    Returns:
        tuple: A tuple containing aggregated results.
        - preds (np.ndarray[N, K, 2]): Predicted keypoint location.
        - maxvals (np.ndarray[N, K, 1]): Scores (confidence) of the keypoints.
    """
    assert isinstance(heatmaps, np.ndarray), "heatmaps should be numpy.ndarray"
    assert heatmaps.ndim == 4, "batch_images should be 4-ndim"

    N, K, _, W = heatmaps.shape
    heatmaps_reshaped = heatmaps.reshape((N, K, -1))
    idx = np.argmax(heatmaps_reshaped, 2).reshape((N, K, 1))
    maxvals = np.amax(heatmaps_reshaped, 2).reshape((N, K, 1))

    preds = np.tile(idx, (1, 1, 2)).astype(np.float32)
    preds[:, :, 0] = preds[:, :, 0] % W
    preds[:, :, 1] = preds[:, :, 1] // W

    preds = np.where(np.tile(maxvals, (1, 1, 2)) > 0.0, preds, -1)
    return preds, maxvals


def transform_preds(coords, center, scale, output_size, use_udp=False):
    """Get final keypoint predictions from heatmaps and apply scaling and
    translation to map them back to the image.
    Note:
        num_keypoints: K
    Args:
        coords (np.ndarray[K, ndims]):
            * If ndims=2, corrds are predicted keypoint location.
            * If ndims=4, corrds are composed of (x, y, scores, tags)
            * If ndims=5, corrds are composed of (x, y, scores, tags,
              flipped_tags)
        center (np.ndarray[2, ]): Center of the bounding box (x, y).
        scale (np.ndarray[2, ]): Scale of the bounding box
            wrt [width, height].
        output_size (np.ndarray[2, ] | list(2,)): Size of the
            destination heatmaps.
        use_udp (bool): Use unbiased data processing
    Returns:
        np.ndarray: Predicted coordinates in the images.
    """
    assert coords.shape[1] in (2, 4, 5)
    assert len(center) == 2
    assert len(scale) == 2
    assert len(output_size) == 2

    # Recover the scale which is normalized by a factor of 200.
    scale = scale * 200.0

    if use_udp:
        scale_x = scale[0] / (output_size[0] - 1.0)
        scale_y = scale[1] / (output_size[1] - 1.0)
    else:
        scale_x = scale[0] / output_size[0]
        scale_y = scale[1] / output_size[1]

    target_coords = np.ones_like(coords)
    target_coords[:, 0] = coords[:, 0] * scale_x + center[0] - scale[0] * 0.5
    target_coords[:, 1] = coords[:, 1] * scale_y + center[1] - scale[1] * 0.5

    return target_coords


def keypoints_from_heatmaps(
    heatmaps,
    center,
    scale,
    unbiased=False,
    post_process="default",
    kernel=11,
    valid_radius_factor=0.0546875,
    use_udp=False,
    target_type="GaussianHeatmap",
):
    """Get final keypoint predictions from heatmaps and transform them back to
    the image.

    Note:
        - batch size: N
        - num keypoints: K
        - heatmap height: H
        - heatmap width: W

    Args:
        heatmaps (np.ndarray[N, K, H, W]): model predicted heatmaps.
        center (np.ndarray[N, 2]): Center of the bounding box (x, y).
        scale (np.ndarray[N, 2]): Scale of the bounding box
            wrt height/width.
        post_process (str/None): Choice of methods to post-process
            heatmaps. Currently supported: None, 'default', 'unbiased',
            'megvii'.
        unbiased (bool): Option to use unbiased decoding. Mutually
            exclusive with megvii.
            Note: this arg is deprecated and unbiased=True can be replaced
            by post_process='unbiased'
            Paper ref: Zhang et al. Distribution-Aware Coordinate
            Representation for Human Pose Estimation (CVPR 2020).
        kernel (int): Gaussian kernel size (K) for modulation, which should
            match the heatmap gaussian sigma when training.
            K=17 for sigma=3 and k=11 for sigma=2.
        valid_radius_factor (float): The radius factor of the positive area
            in classification heatmap for UDP.
        use_udp (bool): Use unbiased data processing.
        target_type (str): 'GaussianHeatmap' or 'CombinedTarget'.
            GaussianHeatmap: Classification target with gaussian distribution.
            CombinedTarget: The combination of classification target
            (response map) and regression target (offset map).
            Paper ref: Huang et al. The Devil is in the Details: Delving into
            Unbiased Data Processing for Human Pose Estimation (CVPR 2020).

    Returns:
        tuple: A tuple containing keypoint predictions and scores.

        - preds (np.ndarray[N, K, 2]): Predicted keypoint location in images.
        - maxvals (np.ndarray[N, K, 1]): Scores (confidence) of the keypoints.
    """
    # Avoid being affected
    heatmaps = heatmaps.copy()

    # detect conflicts
    if unbiased:
        assert post_process not in [False, None, "megvii"]
    if post_process in ["megvii", "unbiased"]:
        assert kernel > 0
    if use_udp:
        assert not post_process == "megvii"

    # normalize configs
    if post_process is False:
        warnings.warn("post_process=False is deprecated, " "please use post_process=None instead", DeprecationWarning)
        post_process = None
    elif post_process is True:
        if unbiased is True:
            warnings.warn(
                "post_process=True, unbiased=True is deprecated," " please use post_process='unbiased' instead",
                DeprecationWarning,
            )
            post_process = "unbiased"
        else:
            warnings.warn(
                "post_process=True, unbiased=False is deprecated, " "please use post_process='default' instead",
                DeprecationWarning,
            )
            post_process = "default"
    elif post_process == "default":
        if unbiased is True:
            warnings.warn(
                "unbiased=True is deprecated, please use " "post_process='unbiased' instead", DeprecationWarning
            )
            post_process = "unbiased"

    # start processing
    if post_process == "megvii":
        heatmaps = _gaussian_blur(heatmaps, kernel=kernel)

    N, K, H, W = heatmaps.shape
    if use_udp:
        if target_type.lower() == "GaussianHeatMap".lower():
            preds, maxvals = _get_max_preds(heatmaps)
            preds = post_dark_udp(preds, heatmaps, kernel=kernel)
        elif target_type.lower() == "CombinedTarget".lower():
            for person_heatmaps in heatmaps:
                for i, heatmap in enumerate(person_heatmaps):
                    kt = 2 * kernel + 1 if i % 3 == 0 else kernel
                    cv2.GaussianBlur(heatmap, (kt, kt), 0, heatmap)
            # valid radius is in direct proportion to the height of heatmap.
            valid_radius = valid_radius_factor * H
            offset_x = heatmaps[:, 1::3, :].flatten() * valid_radius
            offset_y = heatmaps[:, 2::3, :].flatten() * valid_radius
            heatmaps = heatmaps[:, ::3, :]
            preds, maxvals = _get_max_preds(heatmaps)
            index = preds[..., 0] + preds[..., 1] * W
            index += W * H * np.arange(0, N * K / 3)
            index = index.astype(int).reshape(N, K // 3, 1)
            preds += np.concatenate((offset_x[index], offset_y[index]), axis=2)
        else:
            raise ValueError("target_type should be either " "'GaussianHeatmap' or 'CombinedTarget'")
    else:
        preds, maxvals = _get_max_preds(heatmaps)
        if post_process == "unbiased":  # alleviate biased coordinate
            # apply Gaussian distribution modulation.
            heatmaps = np.log(np.maximum(_gaussian_blur(heatmaps, kernel), 1e-10))
            for n in range(N):
                for k in range(K):
                    preds[n][k] = _taylor(heatmaps[n][k], preds[n][k])
        elif post_process is not None:
            # add +/-0.25 shift to the predicted locations for higher acc.
            for n in range(N):
                for k in range(K):
                    heatmap = heatmaps[n][k]
                    px = int(preds[n][k][0])
                    py = int(preds[n][k][1])
                    if 1 < px < W - 1 and 1 < py < H - 1:
                        diff = np.array(
                            [heatmap[py][px + 1] - heatmap[py][px - 1], heatmap[py + 1][px] - heatmap[py - 1][px]]
                        )
                        preds[n][k] += np.sign(diff) * 0.25
                        if post_process == "megvii":
                            preds[n][k] += 0.5

    # Transform back to the image
    for i in range(N):
        preds[i] = transform_preds(preds[i], center[i], scale[i], [W, H], use_udp=use_udp)

    if post_process == "megvii":
        maxvals = maxvals / 255.0 + 0.5

    return preds, maxvals


def preprocess(image):
    mean = (123.675, 116.28, 103.53)
    std = (58.395, 57.120000000000005, 57.375)

    mean = np.array(mean, dtype=np.float32)
    std = np.array(std, dtype=np.float32)
    denominator = np.reciprocal(std, dtype=np.float32)

    image = image - mean
    image = image * denominator
    image = image[np.newaxis,]
    image = np.transpose(image, [0, 3, 1, 2])

    return image



import json
import tempfile
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Type, Union

import fsspec
import numpy as np

from cv_pipeliner.core.inference_model import get_preprocess_input_from_script_file
from cv_pipeliner.inference_models.detection.core import (
    DetectionInput,
    DetectionModel,
    DetectionModelSpec,
    DetectionOutput,
)
from cv_pipeliner.utils.images import (
    denormalize_bboxes,
    rescale_bboxes_with_pad,
    tf_resize_with_pad,
)


class YOLOv5_ModelSpec(DetectionModelSpec):
    """
    note: model_path can be set as torch.hub.load('ultralytics/yolov5', 'yolov5s')
    """

    model_path: Optional[Union[str, Path, "torch.nn.Module"]]  # noqa: F821
    class_names: Optional[Union[List[str], str, Path]] = None
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    device: str = None
    force_reload: bool = False
    skip_validation: bool = False

    @property
    def inference_model_cls(self) -> Type["YOLOv5_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.yolov5 import YOLOv5_DetectionModel

        return YOLOv5_DetectionModel


class YOLOv5_TFLite_ModelSpec(DetectionModelSpec):
    """
    note: model_path can be set as torch.hub.load('ultralytics/yolov5', 'yolov5s')
    """

    model_path: Union[str, Path]
    bboxes_output_index: Union[int, str]
    scores_output_index: Union[int, str]
    classes_output_index: Union[int, str]
    class_names: Optional[Union[List[str], str, Path]] = None
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    use_default_preprocces_and_postprocess_input: bool = False  # (taken from YOLOv5)

    @property
    def inference_model_cls(self) -> Type["YOLOv5_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.yolov5 import YOLOv5_DetectionModel

        return YOLOv5_DetectionModel


class YOLOv5_TFLiteWithNMS_ModelSpec(DetectionModelSpec):
    """
    note: model_path can be set as torch.hub.load('ultralytics/yolov5', 'yolov5s')
    """

    model_path: Union[str, Path]
    class_names: Optional[Union[List[str], str, Path]] = None
    preprocess_input: Union[Callable[[List[np.ndarray]], np.ndarray], str, Path, None] = None
    use_default_preprocces_and_postprocess_input: bool = False  # (taken from YOLOv5)

    @property
    def inference_model_cls(self) -> Type["YOLOv5_DetectionModel"]:
        from cv_pipeliner.inference_models.detection.yolov5 import YOLOv5_DetectionModel

        return YOLOv5_DetectionModel


class YOLOv5_DetectionModel(DetectionModel):
    def _load_yolov5_model(self, model_spec: YOLOv5_ModelSpec):
        import torch

        if isinstance(model_spec.model_path, torch.nn.Module):
            self.model = model_spec.model_path
            return

        temp_file = tempfile.NamedTemporaryFile(suffix=".pt")
        with fsspec.open(model_spec.model_path, "rb") as src:
            temp_file.write(src.read())
        model_path_tmp = Path(temp_file.name)
        self.model = torch.hub.load(
            "ultralytics/yolov5:v7.0",
            "custom",
            path=str(model_path_tmp),
            force_reload=model_spec.force_reload,
            skip_validation=model_spec.skip_validation,
        )
        if model_spec.device is not None:
            self.model = self.model.to(model_spec.device)
        temp_file.close()

    def __init__(self, model_spec: Union[YOLOv5_ModelSpec, YOLOv5_TFLite_ModelSpec, YOLOv5_TFLiteWithNMS_ModelSpec]):
        super().__init__(model_spec)

        if model_spec.class_names is not None:
            if isinstance(model_spec.class_names, str) or isinstance(model_spec.class_names, Path):
                with fsspec.open(model_spec.class_names, "r", encoding="utf-8") as out:
                    self.class_names = np.array(json.load(out))
            else:
                self.class_names = np.array(model_spec.class_names)
        else:
            self.class_names = None

        if isinstance(model_spec.preprocess_input, str) or isinstance(model_spec.preprocess_input, Path):
            self._preprocess_input = get_preprocess_input_from_script_file(script_file=model_spec.preprocess_input)
        else:
            if model_spec.preprocess_input is None:
                self._preprocess_input = lambda x: x
            else:
                self._preprocess_input = model_spec.preprocess_input

        if isinstance(model_spec, YOLOv5_ModelSpec):
            self._load_yolov5_model(model_spec)
            self._raw_predict_images = self._raw_predict_images_torch
        elif isinstance(model_spec, (YOLOv5_TFLite_ModelSpec, YOLOv5_TFLiteWithNMS_ModelSpec)):
            self._load_yolov5_tflite(model_spec)
            if model_spec.use_default_preprocces_and_postprocess_input:
                _, height, width, _ = self.model.get_input_details()[0]["shape"]  # (1 x H x W x 3)
                assert model_spec.preprocess_input is None
                if isinstance(model_spec, YOLOv5_TFLiteWithNMS_ModelSpec):
                    self._preprocess_input = lambda images: [
                        np.array(
                            tf_resize_with_pad(
                                image=image, target_width=width, target_height=height, constant_values=114
                            )
                        )
                        / 255.0
                        for image in images
                    ]
                else:
                    self._preprocess_input = lambda images: [
                        np.array(
                            tf_resize_with_pad(
                                image=image, target_width=width, target_height=height, constant_values=114
                            )
                        )
                        for image in images
                    ]
            self._raw_predict_images = self._raw_predict_images_tflite
        else:
            raise ValueError(f"ObjectDetectionAPI_Model got unknown DetectionModelSpec: {type(model_spec)}")

    def _load_yolov5_tflite(self, model_spec: Union[YOLOv5_TFLite_ModelSpec, YOLOv5_TFLiteWithNMS_ModelSpec]):
        import tensorflow as tf

        temp_file = tempfile.NamedTemporaryFile()
        with fsspec.open(model_spec.model_path, "rb") as src:
            temp_file.write(src.read())
        model_path = Path(temp_file.name)

        self.model = tf.lite.Interpreter(model_path=str(model_path))
        self.model.allocate_tensors()
        self.input_detail = self.model.get_input_details()[0]
        self.input_index = self.input_detail["index"]
        self.input_dtype = self.input_detail["dtype"]
        output_details = self.model.get_output_details()
        output_name_to_index = {output_detail["name"]: output_detail["index"] for output_detail in output_details}
        if isinstance(model_spec, YOLOv5_TFLite_ModelSpec):
            if isinstance(model_spec.bboxes_output_index, str):
                self.bboxes_index = output_name_to_index[model_spec.bboxes_output_index]
            else:
                self.bboxes_index = output_details[model_spec.bboxes_output_index]["index"]
            if isinstance(model_spec.bboxes_output_index, str):
                self.scores_index = output_name_to_index[model_spec.scores_output_index]
            else:
                self.scores_index = output_details[model_spec.scores_output_index]["index"]
            if isinstance(model_spec.classes_output_index, str):
                self.classes_index = output_name_to_index[model_spec.classes_output_index]
            else:
                self.classes_index = output_details[model_spec.classes_output_index]["index"]
        elif isinstance(model_spec, YOLOv5_TFLiteWithNMS_ModelSpec):
            self.output_detail = self.model.get_output_details()[0]["index"]
        temp_file.close()

    def _raw_predict_images_torch(
        self, input: DetectionInput, score_threshold: float
    ) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
        self.model.conf = score_threshold
        results = self.model(input)
        results_pd = results.pandas()

        n_raw_bboxes, n_raw_keypoints, n_raw_scores, n_raw_classes = [], [], [], []
        for result_pd in results_pd.xyxyn:
            n_raw_bboxes.append(np.array(result_pd[["xmin", "ymin", "xmax", "ymax"]]))
            n_raw_keypoints.append(np.array([]).reshape(len(result_pd), 0, 2))
            n_raw_scores.append(np.array(result_pd["confidence"]))
            n_raw_classes.append(np.array(result_pd["class"]))
        return n_raw_bboxes, n_raw_keypoints, n_raw_scores, n_raw_classes

    def _xywh2xyxy_tf(self, xywh: np.ndarray):
        import tensorflow as tf

        x, y, w, h = tf.split(xywh, num_or_size_splits=4, axis=-1)
        return tf.concat([x - w / 2, y - h / 2, x + w / 2, y + h / 2], axis=-1)

    def _post_process_raw_predictions_yolov5(
        self, raw_preds: np.ndarray, max_output_size: int
    ) -> "CombinedNonMaxSuppression":
        import tensorflow as tf

        bboxes = self._xywh2xyxy_tf(raw_preds[..., :4])
        probs = raw_preds[0, :, 4:5]
        classes = raw_preds[0, :, 5:]
        scores = probs * classes
        weights_count = len(classes)
        bboxes = tf.reshape(bboxes, (weights_count, 4))
        scores = tf.reshape(scores, (weights_count,))
        classes = tf.reshape(classes, (weights_count,))

        selected_indices = tf.image.non_max_suppression(
            bboxes,
            scores,
            max_output_size=max_output_size,
            iou_threshold=0.45,
        )
        output_size = len(selected_indices)
        raw_bboxes = tf.gather(bboxes, selected_indices)
        raw_scores = tf.gather(scores, selected_indices)
        raw_classes = tf.gather(classes, selected_indices)

        raw_bboxes = tf.reshape(raw_bboxes, (1, output_size, 4))
        raw_scores = tf.reshape(raw_scores, (1, output_size))
        raw_classes = tf.reshape(raw_classes, (1, output_size))

        raw_bboxes_np = raw_bboxes.numpy()[0]
        raw_scores_np = raw_scores.numpy()[0]
        raw_classes_np = raw_classes.numpy()[0]

        return raw_bboxes_np, raw_scores_np, raw_classes_np

    def _raw_predict_images_tflite(
        self,
        input: DetectionInput,
        score_threshold: float,
    ) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
        n_raw_bboxes, n_raw_keypoints, n_raw_scores, n_raw_classes = [], [], [], []
        for image in input:
            height, width, _ = image.shape
            image = np.array(image[None, ...], dtype=self.input_dtype)
            self.model.resize_tensor_input(0, [1, height, width, 3])
            self.model.allocate_tensors()
            self.model.set_tensor(self.input_index, image)
            self.model.invoke()
            if isinstance(self.model_spec, YOLOv5_TFLite_ModelSpec):
                raw_bboxes = np.array(self.model.get_tensor(self.bboxes_index))[0]
                raw_scores = np.array(self.model.get_tensor(self.scores_index))[0]
                raw_classes = np.array(self.model.get_tensor(self.classes_index))[0]
            else:
                raw_preds = np.array(self.model.get_tensor(self.output_detail))
                raw_bboxes, raw_scores, raw_classes = self._post_process_raw_predictions_yolov5(
                    raw_preds, max_output_size=2000
                )
            n_raw_bboxes.append(raw_bboxes)
            n_raw_keypoints.append(np.array([]).reshape(len(raw_bboxes), 0, 2))
            n_raw_scores.append(raw_scores)
            n_raw_classes.append(raw_classes)

        return n_raw_bboxes, n_raw_keypoints, n_raw_scores, n_raw_classes

    def _postprocess_prediction(
        self,
        raw_bboxes: np.ndarray,
        raw_keypoints: np.ndarray,
        raw_scores: np.ndarray,
        raw_classes: np.ndarray,
        score_threshold: float,
        current_width: int,
        current_height: int,
        target_width: int,
        target_height: int,
        classification_top_n: int,
    ) -> Tuple[
        List[Tuple[int, int, int, int]], List[List[Tuple[int, int]]], List[float], List[List[str]], List[List[float]]
    ]:
        if isinstance(self.model_spec, (YOLOv5_TFLite_ModelSpec, YOLOv5_TFLiteWithNMS_ModelSpec)) and (
            self.model_spec.use_default_preprocces_and_postprocess_input
        ):
            # Rescale bboxes
            raw_bboxes = denormalize_bboxes(raw_bboxes, current_width, current_height)
            raw_bboxes = rescale_bboxes_with_pad(
                bboxes=raw_bboxes,
                current_width=current_width,
                current_height=current_height,
                target_width=target_width,
                target_height=target_height,
            )
        else:
            raw_bboxes = denormalize_bboxes(raw_bboxes, target_width, target_height)

        mask = raw_scores > score_threshold
        bboxes = raw_bboxes[mask]
        keypoints = raw_keypoints[mask]
        scores = raw_scores[mask]
        classes = raw_classes[mask]

        correct_non_repeated_bboxes_idxs = []
        bboxes_set = set()
        for idx, bbox in enumerate(bboxes):
            xmin, ymin, xmax, ymax = bbox
            if xmax - xmin > 0 and ymax - ymin > 0 and (xmin, ymin, xmax, ymax) not in bboxes_set:
                bboxes_set.add((xmin, ymin, xmax, ymax))
                correct_non_repeated_bboxes_idxs.append(idx)

        bboxes = bboxes[correct_non_repeated_bboxes_idxs]
        keypoints = keypoints[correct_non_repeated_bboxes_idxs]
        scores = scores[correct_non_repeated_bboxes_idxs]
        classes = classes[correct_non_repeated_bboxes_idxs]
        classes_scores = scores.copy()
        if self.class_names is not None:
            if classification_top_n > 1:
                raise NotImplementedError("Not impelemented for classification_top_n > 1")
            class_names_top_n = np.array(
                [
                    [class_name for i in range(classification_top_n)]
                    for class_name in self.class_names[classes.astype(np.int32)]
                ]
            )
            classes_scores_top_n = np.array(
                [[[score for _ in range(classification_top_n)]] for score in classes_scores]
            )
        else:
            class_names_top_n = np.array([[None for _ in range(classification_top_n)] for _ in classes])
            classes_scores_top_n = np.array([[score for _ in range(classification_top_n)] for score in classes_scores])
        return bboxes, keypoints, scores, class_names_top_n, classes_scores_top_n

    def predict(self, input: DetectionInput, score_threshold: float, classification_top_n: int = 1) -> DetectionOutput:
        target_heights_widths = [image.shape[:2] for image in input]
        input = self.preprocess_input(input)
        current_heights_widths = [image.shape[:2] for image in input]
        n_raw_bboxes, n_raw_keypoints, n_raw_scores, n_raw_classes = self._raw_predict_images(input, score_threshold)
        results = [
            self._postprocess_prediction(
                raw_bboxes=raw_bboxes,
                raw_keypoints=raw_keypoints,
                raw_scores=raw_scores,
                raw_classes=raw_classes,
                score_threshold=score_threshold,
                current_height=current_height,
                current_width=current_width,
                target_width=target_width,
                target_height=target_height,
                classification_top_n=classification_top_n,
            )
            for (
                (target_height, target_width),
                (current_height, current_width),
                raw_bboxes,
                raw_keypoints,
                raw_scores,
                raw_classes,
            ) in zip(
                target_heights_widths,
                current_heights_widths,
                n_raw_bboxes,
                n_raw_keypoints,
                n_raw_scores,
                n_raw_classes,
            )
        ]
        n_pred_bboxes, n_pred_keypoints, n_pred_scores, n_pred_class_names_top_k, n_pred_scores_top_k = [
            [res[i] for res in results] for i in range(5)
        ]
        n_pred_masks = [[[] for _ in pred_bboxes] for pred_bboxes in n_pred_bboxes]
        return (
            n_pred_bboxes,
            n_pred_keypoints,
            n_pred_masks,
            n_pred_scores,
            n_pred_class_names_top_k,
            n_pred_scores_top_k,
        )

    def preprocess_input(self, input: DetectionInput):
        return self._preprocess_input(input)

    @property
    def input_size(self) -> Tuple[int, int]:
        return self.model_spec.input_size
